@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2018 Red Hat Inc.
+ * Copyright (C) 2018-2019 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -44,15 +44,20 @@
 
 #include <nbdkit-filter.h>
 
-#include "blk.h"
+#include "cleanup.h"
 
-#define THREAD_MODEL NBDKIT_THREAD_MODEL_PARALLEL
+#include "blk.h"
+#include "isaligned.h"
+#include "minmax.h"
+#include "rounding.h"
 
 /* In order to handle parallel requests safely, this lock must be held
  * when calling any blk_* functions.
  */
 static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
 
+bool cow_on_cache;
+
 static void
 cow_load (void)
 {
@@ -66,6 +71,27 @@ cow_unload (void)
   blk_free ();
 }
 
+static int
+cow_config (nbdkit_next_config *next, void *nxdata,
+            const char *key, const char *value)
+{
+  if (strcmp (key, "cow-on-cache") == 0) {
+    int r;
+
+    r = nbdkit_parse_bool (value);
+    if (r == -1)
+      return -1;
+    cow_on_cache = r;
+    return 0;
+  }
+  else {
+    return next (nxdata, key, value);
+  }
+}
+
+#define cow_config_help \
+  "cow-on-cache=<BOOL>  Set to true to treat client cache requests as writes.\n"
+
 static void *
 cow_open (nbdkit_next_open *next, void *nxdata, int readonly)
 {
@@ -76,7 +102,9 @@ cow_open (nbdkit_next_open *next, void *nxdata, int readonly)
   return NBDKIT_HANDLE_NOT_NEEDED;
 }
 
-/* Get the file size and ensure the overlay is the correct size. */
+/* Get the file size; round it down to overlay granularity before
+ * setting overlay size.
+ */
 static int64_t
 cow_get_size (struct nbdkit_next_ops *next_ops, void *nxdata,
               void *handle)
@@ -89,10 +117,10 @@ cow_get_size (struct nbdkit_next_ops *next_ops, void *nxdata,
     return -1;
 
   nbdkit_debug ("cow: underlying file size: %" PRIi64, size);
+  size = ROUND_DOWN (size, BLKSIZE);
 
-  pthread_mutex_lock (&lock);
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
   r = blk_set_size (size);
-  pthread_mutex_unlock (&lock);
   if (r == -1)
     return -1;
 
@@ -104,7 +132,7 @@ cow_get_size (struct nbdkit_next_ops *next_ops, void *nxdata,
  */
 static int
 cow_prepare (struct nbdkit_next_ops *next_ops, void *nxdata,
-             void *handle)
+             void *handle, int readonly)
 {
   int64_t r;
 
@@ -145,6 +173,29 @@ cow_can_fua (struct nbdkit_next_ops *next_ops, void *nxdata, void *handle)
   return NBDKIT_FUA_EMULATE;
 }
 
+static int
+cow_can_cache (struct nbdkit_next_ops *next_ops, void *nxdata, void *handle)
+{
+  /* Cache next_ops->can_cache now, so later calls to next_ops->cache
+   * don't fail, even though we override the answer here.
+   */
+  int r = next_ops->can_cache (nxdata);
+  if (r == -1)
+    return -1;
+  return NBDKIT_FUA_NATIVE;
+}
+
+/* Override the plugin's .can_fast_zero, because our .zero is not fast */
+static int
+cow_can_fast_zero (struct nbdkit_next_ops *next_ops, void *nxdata,
+                   void *handle)
+{
+  /* It is better to advertise support even when we always reject fast
+   * zero attempts.
+   */
+  return 1;
+}
+
 static int cow_flush (struct nbdkit_next_ops *next_ops, void *nxdata, void *handle, uint32_t flags, int *err);
 
 /* Read data. */
@@ -153,41 +204,70 @@ cow_pread (struct nbdkit_next_ops *next_ops, void *nxdata,
            void *handle, void *buf, uint32_t count, uint64_t offset,
            uint32_t flags, int *err)
 {
-  uint8_t *block;
+  CLEANUP_FREE uint8_t *block = NULL;
+  uint64_t blknum, blkoffs;
+  int r;
 
-  block = malloc (BLKSIZE);
-  if (block == NULL) {
-    *err = errno;
-    nbdkit_error ("malloc: %m");
-    return -1;
-  }
-
-  while (count > 0) {
-    uint64_t blknum, blkoffs, n;
-    int r;
-
-    blknum = offset / BLKSIZE;  /* block number */
-    blkoffs = offset % BLKSIZE; /* offset within the block */
-    n = BLKSIZE - blkoffs;      /* max bytes we can read from this block */
-    if (n > count)
-      n = count;
-
-    pthread_mutex_lock (&lock);
-    r = blk_read (next_ops, nxdata, blknum, block, err);
-    pthread_mutex_unlock (&lock);
-    if (r == -1) {
-      free (block);
+  if (!IS_ALIGNED (count | offset, BLKSIZE)) {
+    block = malloc (BLKSIZE);
+    if (block == NULL) {
+      *err = errno;
+      nbdkit_error ("malloc: %m");
       return -1;
     }
+  }
+
+  blknum = offset / BLKSIZE;  /* block number */
+  blkoffs = offset % BLKSIZE; /* offset within the block */
+
+  /* Unaligned head */
+  if (blkoffs) {
+    uint64_t n = MIN (BLKSIZE - blkoffs, count);
+
+    assert (block);
+    ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
+    r = blk_read (next_ops, nxdata, blknum, block, err);
+    if (r == -1)
+      return -1;
 
     memcpy (buf, &block[blkoffs], n);
 
     buf += n;
     count -= n;
     offset += n;
+    blknum++;
+  }
+
+  /* Aligned body */
+  /* XXX This breaks up large read requests into smaller ones, which
+   * is a problem for plugins which have a large, fixed per-request
+   * overhead (hello, curl).  We should try to keep large requests
+   * together as much as possible, but that requires us to be much
+   * smarter here.
+   */
+  while (count >= BLKSIZE) {
+    ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
+    r = blk_read (next_ops, nxdata, blknum, buf, err);
+    if (r == -1)
+      return -1;
+
+    buf += BLKSIZE;
+    count -= BLKSIZE;
+    offset += BLKSIZE;
+    blknum++;
+  }
+
+  /* Unaligned tail */
+  if (count) {
+    assert (block);
+    ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
+    r = blk_read (next_ops, nxdata, blknum, block, err);
+    if (r == -1)
+      return -1;
+
+    memcpy (buf, block, count);
   }
 
-  free (block);
   return 0;
 }
 
@@ -197,46 +277,71 @@ cow_pwrite (struct nbdkit_next_ops *next_ops, void *nxdata,
             void *handle, const void *buf, uint32_t count, uint64_t offset,
             uint32_t flags, int *err)
 {
-  uint8_t *block;
+  CLEANUP_FREE uint8_t *block = NULL;
+  uint64_t blknum, blkoffs;
+  int r;
 
-  block = malloc (BLKSIZE);
-  if (block == NULL) {
-    *err = errno;
-    nbdkit_error ("malloc: %m");
-    return -1;
+  if (!IS_ALIGNED (count | offset, BLKSIZE)) {
+    block = malloc (BLKSIZE);
+    if (block == NULL) {
+      *err = errno;
+      nbdkit_error ("malloc: %m");
+      return -1;
+    }
   }
 
-  while (count > 0) {
-    uint64_t blknum, blkoffs, n;
-    int r;
+  blknum = offset / BLKSIZE;  /* block number */
+  blkoffs = offset % BLKSIZE; /* offset within the block */
 
-    blknum = offset / BLKSIZE;  /* block number */
-    blkoffs = offset % BLKSIZE; /* offset within the block */
-    n = BLKSIZE - blkoffs;      /* max bytes we can read from this block */
-    if (n > count)
-      n = count;
+  /* Unaligned head */
+  if (blkoffs) {
+    uint64_t n = MIN (BLKSIZE - blkoffs, count);
 
     /* Do a read-modify-write operation on the current block.
      * Hold the lock over the whole operation.
      */
-    pthread_mutex_lock (&lock);
+    assert (block);
+    ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
     r = blk_read (next_ops, nxdata, blknum, block, err);
     if (r != -1) {
       memcpy (&block[blkoffs], buf, n);
       r = blk_write (blknum, block, err);
     }
-    pthread_mutex_unlock (&lock);
-    if (r == -1) {
-      free (block);
+    if (r == -1)
       return -1;
-    }
 
     buf += n;
     count -= n;
     offset += n;
+    blknum++;
+  }
+
+  /* Aligned body */
+  while (count >= BLKSIZE) {
+    ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
+    r = blk_write (blknum, buf, err);
+    if (r == -1)
+      return -1;
+
+    buf += BLKSIZE;
+    count -= BLKSIZE;
+    offset += BLKSIZE;
+    blknum++;
+  }
+
+  /* Unaligned tail */
+  if (count) {
+    assert (block);
+    ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
+    r = blk_read (next_ops, nxdata, blknum, block, err);
+    if (r != -1) {
+      memcpy (block, buf, count);
+      r = blk_write (blknum, block, err);
+    }
+    if (r == -1)
+      return -1;
   }
 
-  free (block);
   if (flags & NBDKIT_FLAG_FUA)
     return cow_flush (next_ops, nxdata, handle, 0, err);
   return 0;
@@ -248,7 +353,17 @@ cow_zero (struct nbdkit_next_ops *next_ops, void *nxdata,
           void *handle, uint32_t count, uint64_t offset, uint32_t flags,
           int *err)
 {
-  uint8_t *block;
+  CLEANUP_FREE uint8_t *block = NULL;
+  uint64_t blknum, blkoffs;
+  int r;
+
+  /* We are purposefully avoiding next_ops->zero, so a zero request is
+   * never faster than plain writes.
+   */
+  if (flags & NBDKIT_FLAG_FAST_ZERO) {
+    *err = ENOTSUP;
+    return -1;
+  }
 
   block = malloc (BLKSIZE);
   if (block == NULL) {
@@ -257,39 +372,59 @@ cow_zero (struct nbdkit_next_ops *next_ops, void *nxdata,
     return -1;
   }
 
-  while (count > 0) {
-    uint64_t blknum, blkoffs, n;
-    int r;
+  blknum = offset / BLKSIZE;  /* block number */
+  blkoffs = offset % BLKSIZE; /* offset within the block */
 
-    blknum = offset / BLKSIZE;  /* block number */
-    blkoffs = offset % BLKSIZE; /* offset within the block */
-    n = BLKSIZE - blkoffs;      /* max bytes we can read from this block */
-    if (n > count)
-      n = count;
+  /* Unaligned head */
+  if (blkoffs) {
+    uint64_t n = MIN (BLKSIZE - blkoffs, count);
 
     /* Do a read-modify-write operation on the current block.
      * Hold the lock over the whole operation.
-     *
-     * XXX There is the possibility of optimizing this: ONLY if we are
-     * writing a whole, aligned block, then use FALLOC_FL_ZERO_RANGE.
      */
-    pthread_mutex_lock (&lock);
+    ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
     r = blk_read (next_ops, nxdata, blknum, block, err);
     if (r != -1) {
       memset (&block[blkoffs], 0, n);
       r = blk_write (blknum, block, err);
     }
-    pthread_mutex_unlock (&lock);
-    if (r == -1) {
-      free (block);
+    if (r == -1)
       return -1;
-    }
 
     count -= n;
     offset += n;
+    blknum++;
+  }
+
+  /* Aligned body */
+  if (count >= BLKSIZE)
+    memset (block, 0, BLKSIZE);
+  while (count >= BLKSIZE) {
+    /* XXX There is the possibility of optimizing this: since this loop is
+     * writing a whole, aligned block, we should use FALLOC_FL_ZERO_RANGE.
+     */
+    ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
+    r = blk_write (blknum, block, err);
+    if (r == -1)
+      return -1;
+
+    count -= BLKSIZE;
+    offset += BLKSIZE;
+    blknum++;
+  }
+
+  /* Unaligned tail */
+  if (count) {
+    ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
+    r = blk_read (next_ops, nxdata, blknum, block, err);
+    if (r != -1) {
+      memset (&block[count], 0, BLKSIZE - count);
+      r = blk_write (blknum, block, err);
+    }
+    if (r == -1)
+      return -1;
   }
 
-  free (block);
   if (flags & NBDKIT_FLAG_FUA)
     return cow_flush (next_ops, nxdata, handle, 0, err);
   return 0;
@@ -301,21 +436,81 @@ cow_flush (struct nbdkit_next_ops *next_ops, void *nxdata, void *handle,
 {
   int r;
 
-  pthread_mutex_lock (&lock);
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
   r = blk_flush ();
   if (r == -1)
     *err = errno;
-  pthread_mutex_unlock (&lock);
   return r;
 }
 
+static int
+cow_cache (struct nbdkit_next_ops *next_ops, void *nxdata,
+           void *handle, uint32_t count, uint64_t offset,
+           uint32_t flags, int *err)
+{
+  CLEANUP_FREE uint8_t *block = NULL;
+  uint64_t blknum, blkoffs;
+  int r;
+  uint64_t remaining = count; /* Rounding out could exceed 32 bits */
+  enum cache_mode mode;
+
+  switch (next_ops->can_cache (nxdata)) {
+  case NBDKIT_CACHE_NONE:
+    mode = BLK_CACHE_IGNORE;
+    break;
+  case NBDKIT_CACHE_EMULATE:
+    mode = BLK_CACHE_READ;
+    break;
+  case NBDKIT_CACHE_NATIVE:
+    mode = BLK_CACHE_PASSTHROUGH;
+    break;
+  default:
+    assert (false); /* Guaranteed thanks to early caching */
+  }
+  if (cow_on_cache)
+    mode = BLK_CACHE_COW;
+
+  assert (!flags);
+  block = malloc (BLKSIZE);
+  if (block == NULL) {
+    *err = errno;
+    nbdkit_error ("malloc: %m");
+    return -1;
+  }
+
+  blknum = offset / BLKSIZE;  /* block number */
+  blkoffs = offset % BLKSIZE; /* offset within the block */
+
+  /* Unaligned head */
+  remaining += blkoffs;
+  offset -= blkoffs;
+
+  /* Unaligned tail */
+  remaining = ROUND_UP (remaining, BLKSIZE);
+
+  /* Aligned body */
+  while (remaining) {
+    ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
+    r = blk_cache (next_ops, nxdata, blknum, block, mode, err);
+    if (r == -1)
+      return -1;
+
+    remaining -= BLKSIZE;
+    offset += BLKSIZE;
+    blknum++;
+  }
+
+  return 0;
+}
+
 static struct nbdkit_filter filter = {
   .name              = "cow",
   .longname          = "nbdkit copy-on-write (COW) filter",
-  .version           = PACKAGE_VERSION,
   .load              = cow_load,
   .unload            = cow_unload,
   .open              = cow_open,
+  .config            = cow_config,
+  .config_help       = cow_config_help,
   .prepare           = cow_prepare,
   .get_size          = cow_get_size,
   .can_write         = cow_can_write,
@@ -323,10 +518,13 @@ static struct nbdkit_filter filter = {
   .can_trim          = cow_can_trim,
   .can_extents       = cow_can_extents,
   .can_fua           = cow_can_fua,
+  .can_cache         = cow_can_cache,
+  .can_fast_zero     = cow_can_fast_zero,
   .pread             = cow_pread,
   .pwrite            = cow_pwrite,
   .zero              = cow_zero,
   .flush             = cow_flush,
+  .cache             = cow_cache,
 };
 
 NBDKIT_REGISTER_FILTER(filter)
