@@ -41,8 +41,6 @@
 
 #include <nbdkit-filter.h>
 
-#define THREAD_MODEL NBDKIT_THREAD_MODEL_PARALLEL
-
 static enum FuaMode {
   NONE,
   EMULATE,
@@ -77,10 +75,15 @@ fua_config (nbdkit_next_config *next, void *nxdata,
 
 /* Check that desired mode is supported by plugin. */
 static int
-fua_prepare (struct nbdkit_next_ops *next_ops, void *nxdata, void *handle)
+fua_prepare (struct nbdkit_next_ops *next_ops, void *nxdata, void *handle,
+             int readonly)
 {
   int r;
 
+  /* If we are opened readonly, this filter has no impact */
+  if (readonly)
+    return 0;
+
   switch (fuamode) {
   case NONE:
     break;
@@ -235,7 +238,6 @@ fua_zero (struct nbdkit_next_ops *next_ops, void *nxdata,
 static struct nbdkit_filter filter = {
   .name              = "fua",
   .longname          = "nbdkit fua filter",
-  .version           = PACKAGE_VERSION,
   .config            = fua_config,
   .config_help       = fua_config_help,
   .prepare           = fua_prepare,
