@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2018 Red Hat Inc.
+ * Copyright (C) 2018-2019 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -42,10 +42,12 @@
 #include <pthread.h>
 #include <sys/time.h>
 #include <assert.h>
+#include <fcntl.h>
+#include <unistd.h>
 
 #include <nbdkit-filter.h>
 
-#define THREAD_MODEL NBDKIT_THREAD_MODEL_PARALLEL
+#include "cleanup.h"
 
 static uint64_t connections;
 static char *logfilename;
@@ -88,13 +90,29 @@ log_config (nbdkit_next_config *next, void *nxdata,
 static int
 log_config_complete (nbdkit_next_config_complete *next, void *nxdata)
 {
+  int fd;
+
   if (!logfilename) {
     nbdkit_error ("missing logfile= parameter for the log filter");
     return -1;
   }
-  logfile = fopen (logfilename, append ? "a" : "w");
+  /* Using fopen("ae"/"we") would be more convenient, but as Haiku
+   * still lacks that, use this instead. Atomicity is not essential
+   * here since .config completes before threads that might fork, if
+   * we have to later add yet another fallback to fcntl(fileno()) for
+   * systems without O_CLOEXEC.
+   */
+  fd = open (logfilename,
+             O_CLOEXEC | O_WRONLY | O_CREAT | (append ? O_APPEND : O_TRUNC),
+             0666);
+  if (fd < 0) {
+    nbdkit_error ("open: %s: %m", logfilename);
+    return -1;
+  }
+  logfile = fdopen (fd, append ? "a" : "w");
   if (!logfile) {
-    nbdkit_error ("fopen: %m");
+    nbdkit_error ("fdopen: %s: %m", logfilename);
+    close (fd);
     return -1;
   }
 
@@ -114,12 +132,8 @@ struct handle {
 static uint64_t
 get_id (struct handle *h)
 {
-  uint64_t r;
-
-  pthread_mutex_lock (&lock);
-  r = ++h->id;
-  pthread_mutex_unlock (&lock);
-  return r;
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE(&lock);
+  return ++h->id;
 }
 
 /* Output a timestamp and the log message. */
@@ -219,9 +233,8 @@ log_open (nbdkit_next_open *next, void *nxdata, int readonly)
     return NULL;
   }
 
-  pthread_mutex_lock (&lock);
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
   h->connection = ++connections;
-  pthread_mutex_unlock (&lock);
   h->id = 0;
   return h;
 }
@@ -235,7 +248,8 @@ log_close (void *handle)
 }
 
 static int
-log_prepare (struct nbdkit_next_ops *next_ops, void *nxdata, void *handle)
+log_prepare (struct nbdkit_next_ops *next_ops, void *nxdata, void *handle,
+             int readonly)
 {
   struct handle *h = handle;
   int64_t size = next_ops->get_size (nxdata);
@@ -245,12 +259,17 @@ log_prepare (struct nbdkit_next_ops *next_ops, void *nxdata, void *handle)
   int t = next_ops->can_trim (nxdata);
   int z = next_ops->can_zero (nxdata);
   int F = next_ops->can_fua (nxdata);
+  int e = next_ops->can_extents (nxdata);
+  int c = next_ops->can_cache (nxdata);
+  int Z = next_ops->can_fast_zero (nxdata);
 
-  if (size < 0 || w < 0 || f < 0 || r < 0 || t < 0 || z < 0 || F < 0)
+  if (size < 0 || w < 0 || f < 0 || r < 0 || t < 0 || z < 0 || F < 0 ||
+      e < 0 || c < 0 || Z < 0)
     return -1;
 
   output (h, "Connect", 0, "size=0x%" PRIx64 " write=%d flush=%d "
-          "rotational=%d trim=%d zero=%d fua=%d", size, w, f, r, t, z, F);
+          "rotational=%d trim=%d zero=%d fua=%d extents=%d cache=%d "
+          "fast_zero=%d", size, w, f, r, t, z, F, e, c, Z);
   return 0;
 }
 
@@ -343,10 +362,13 @@ log_zero (struct nbdkit_next_ops *next_ops, void *nxdata,
   uint64_t id = get_id (h);
   int r;
 
-  assert (!(flags & ~(NBDKIT_FLAG_FUA | NBDKIT_FLAG_MAY_TRIM)));
-  output (h, "Zero", id, "offset=0x%" PRIx64 " count=0x%x trim=%d fua=%d ...",
+  assert (!(flags & ~(NBDKIT_FLAG_FUA | NBDKIT_FLAG_MAY_TRIM |
+                      NBDKIT_FLAG_FAST_ZERO)));
+  output (h, "Zero", id,
+          "offset=0x%" PRIx64 " count=0x%x trim=%d fua=%d fast=%d...",
           offs, count, !!(flags & NBDKIT_FLAG_MAY_TRIM),
-          !!(flags & NBDKIT_FLAG_FUA));
+          !!(flags & NBDKIT_FLAG_FUA),
+          !!(flags & NBDKIT_FLAG_FAST_ZERO));
   r = next_ops->zero (nxdata, count, offs, flags, err);
   output_return (h, "...Zero", id, r, err);
   return r;
@@ -371,7 +393,7 @@ log_extents (struct nbdkit_next_ops *next_ops, void *nxdata,
     output_return (h, "...Extents", id, r, err);
   else {
     FILE *fp;
-    char *extents_str = NULL;
+    CLEANUP_FREE char *extents_str = NULL;
     size_t i, n, len = 0;
 
     fp = open_memstream (&extents_str, &len);
@@ -393,15 +415,31 @@ log_extents (struct nbdkit_next_ops *next_ops, void *nxdata,
 
     output (h, "...Extents", id, "extents=[%s] return=0",
             extents_str ? extents_str : "(null)");
-    free (extents_str);
   }
   return r;
 }
 
+/* Cache data. */
+static int
+log_cache (struct nbdkit_next_ops *next_ops, void *nxdata,
+           void *handle, uint32_t count, uint64_t offs, uint32_t flags,
+           int *err)
+{
+  struct handle *h = handle;
+  uint64_t id = get_id (h);
+  int r;
+
+  assert (!flags);
+  output (h, "Cache", id, "offset=0x%" PRIx64 " count=0x%x ...",
+          offs, count);
+  r = next_ops->cache (nxdata, count, offs, flags, err);
+  output_return (h, "...Cache", id, r, err);
+  return r;
+}
+
 static struct nbdkit_filter filter = {
   .name              = "log",
   .longname          = "nbdkit log filter",
-  .version           = PACKAGE_VERSION,
   .config            = log_config,
   .config_complete   = log_config_complete,
   .config_help       = log_config_help,
@@ -416,6 +454,7 @@ static struct nbdkit_filter filter = {
   .trim              = log_trim,
   .zero              = log_zero,
   .extents           = log_extents,
+  .cache             = log_cache,
 };
 
 NBDKIT_REGISTER_FILTER(filter)
