@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2018 Red Hat Inc.
+ * Copyright (C) 2018-2019 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -39,7 +39,7 @@
 
 #include <nbdkit-filter.h>
 
-#define THREAD_MODEL NBDKIT_THREAD_MODEL_PARALLEL
+#include "cleanup.h"
 
 static int64_t offset = 0, range = -1;
 
@@ -64,16 +64,9 @@ offset_config (nbdkit_next_config *next, void *nxdata,
     return next (nxdata, key, value);
 }
 
-/* Check the user did pass both parameters. */
-static int
-offset_config_complete (nbdkit_next_config_complete *next, void *nxdata)
-{
-  return next (nxdata);
-}
-
 #define offset_config_help \
-  "offset=<OFFSET>     (required) The start offset to serve.\n" \
-  "range=<LENGTH>                 The total size to serve."
+  "offset=<OFFSET>            The start offset to serve (default 0).\n" \
+  "range=<LENGTH>             The total size to serve (default rest of file)."
 
 /* Get the file size. */
 static int64_t
@@ -82,14 +75,22 @@ offset_get_size (struct nbdkit_next_ops *next_ops, void *nxdata,
 {
   int64_t real_size = next_ops->get_size (nxdata);
 
+  if (real_size == -1)
+    return -1;
+
   if (range >= 0) {
-    if (offset + range > real_size) {
+    if (offset > real_size - range) {
       nbdkit_error ("offset+range is larger than the real size "
                     "of the underlying file or device");
       return -1;
     }
     return range;
   }
+  else if (offset > real_size) {
+    nbdkit_error ("offset is larger than the real size "
+                  "of the underlying file or device");
+    return -1;
+  }
   else
     return real_size - offset;
 }
@@ -138,39 +139,43 @@ offset_extents (struct nbdkit_next_ops *next_ops, void *nxdata,
                 struct nbdkit_extents *extents, int *err)
 {
   size_t i;
-  struct nbdkit_extents *extents2;
+  CLEANUP_EXTENTS_FREE struct nbdkit_extents *extents2 = NULL;
   struct nbdkit_extent e;
-  int64_t real_size = next_ops->get_size (nxdata);
+  int64_t end = range >= 0 ? offset + range : next_ops->get_size (nxdata);
 
-  extents2 = nbdkit_extents_new (offs + offset, real_size - offset);
+  extents2 = nbdkit_extents_new (offs + offset, end);
   if (extents2 == NULL) {
     *err = errno;
     return -1;
   }
   if (next_ops->extents (nxdata, count, offs + offset,
                          flags, extents2, err) == -1)
-    goto error;
+    return -1;
 
   for (i = 0; i < nbdkit_extents_count (extents2); ++i) {
     e = nbdkit_get_extent (extents2, i);
     e.offset -= offset;
-    if (nbdkit_add_extent (extents, e.offset, e.length, e.type) == -1)
-      goto error;
+    if (nbdkit_add_extent (extents, e.offset, e.length, e.type) == -1) {
+      *err = errno;
+      return -1;
+    }
   }
-  nbdkit_extents_free (extents2);
   return 0;
+}
 
- error:
-  nbdkit_extents_free (extents2);
-  return -1;
+/* Cache data. */
+static int
+offset_cache (struct nbdkit_next_ops *next_ops, void *nxdata,
+              void *handle, uint32_t count, uint64_t offs, uint32_t flags,
+              int *err)
+{
+  return next_ops->cache (nxdata, count, offs + offset, flags, err);
 }
 
 static struct nbdkit_filter filter = {
   .name              = "offset",
   .longname          = "nbdkit offset filter",
-  .version           = PACKAGE_VERSION,
   .config            = offset_config,
-  .config_complete   = offset_config_complete,
   .config_help       = offset_config_help,
   .get_size          = offset_get_size,
   .pread             = offset_pread,
@@ -178,6 +183,7 @@ static struct nbdkit_filter filter = {
   .trim              = offset_trim,
   .zero              = offset_zero,
   .extents           = offset_extents,
+  .cache             = offset_cache,
 };
 
 NBDKIT_REGISTER_FILTER(filter)
