@@ -41,12 +41,11 @@
 #include <nbdkit-filter.h>
 
 #include "byte-swapping.h"
+#include "cleanup.h"
 
 #include "partition.h"
 
-#define THREAD_MODEL NBDKIT_THREAD_MODEL_PARALLEL
-
-int partnum = -1;
+unsigned partnum = 0;
 
 /* Called for each key=value passed on the command line. */
 static int
@@ -54,7 +53,9 @@ partition_config (nbdkit_next_config *next, void *nxdata,
                   const char *key, const char *value)
 {
   if (strcmp (key, "partition") == 0) {
-    if (sscanf (value, "%d", &partnum) != 1 || partnum <= 0) {
+    if (nbdkit_parse_unsigned ("partition", value, &partnum) == -1)
+      return -1;
+    if (partnum == 0) {
       nbdkit_error ("invalid partition number");
       return -1;
     }
@@ -68,7 +69,7 @@ partition_config (nbdkit_next_config *next, void *nxdata,
 static int
 partition_config_complete (nbdkit_next_config_complete *next, void *nxdata)
 {
-  if (partnum == -1) {
+  if (partnum == 0) {
     nbdkit_error ("you must supply the partition parameter on the command line");
     return -1;
   }
@@ -114,7 +115,7 @@ partition_close (void *handle)
 
 static int
 partition_prepare (struct nbdkit_next_ops *next_ops, void *nxdata,
-                   void *handle)
+                   void *handle, int readonly)
 {
   struct handle *h = handle;
   int64_t size;
@@ -229,37 +230,43 @@ partition_extents (struct nbdkit_next_ops *next_ops, void *nxdata,
 {
   struct handle *h = handle;
   size_t i;
-  struct nbdkit_extents *extents2;
+  CLEANUP_EXTENTS_FREE struct nbdkit_extents *extents2 = NULL;
   struct nbdkit_extent e;
-  int64_t real_size = next_ops->get_size (nxdata);
 
-  extents2 = nbdkit_extents_new (offs + h->offset, real_size - h->offset);
+  extents2 = nbdkit_extents_new (offs + h->offset, h->offset + h->range);
   if (extents2 == NULL) {
     *err = errno;
     return -1;
   }
   if (next_ops->extents (nxdata, count, offs + h->offset,
                          flags, extents2, err) == -1)
-    goto error;
+    return -1;
 
   for (i = 0; i < nbdkit_extents_count (extents2); ++i) {
     e = nbdkit_get_extent (extents2, i);
     e.offset -= h->offset;
-    if (nbdkit_add_extent (extents, e.offset, e.length, e.type) == -1)
-      goto error;
+    if (nbdkit_add_extent (extents, e.offset, e.length, e.type) == -1) {
+      *err = errno;
+      return -1;
+    }
   }
-  nbdkit_extents_free (extents2);
   return 0;
+}
 
- error:
-  nbdkit_extents_free (extents2);
-  return -1;
+/* Cache data. */
+static int
+partition_cache (struct nbdkit_next_ops *next_ops, void *nxdata,
+                 void *handle, uint32_t count, uint64_t offs, uint32_t flags,
+                 int *err)
+{
+  struct handle *h = handle;
+
+  return next_ops->cache (nxdata, count, offs + h->offset, flags, err);
 }
 
 static struct nbdkit_filter filter = {
   .name              = "partition",
   .longname          = "nbdkit partition filter",
-  .version           = PACKAGE_VERSION,
   .config            = partition_config,
   .config_complete   = partition_config_complete,
   .config_help       = partition_config_help,
@@ -272,6 +279,7 @@ static struct nbdkit_filter filter = {
   .trim              = partition_trim,
   .zero              = partition_zero,
   .extents           = partition_extents,
+  .cache             = partition_cache,
 };
 
 NBDKIT_REGISTER_FILTER(filter)
