@@ -48,11 +48,9 @@
 
 #include <lzma.h>
 
-#include "xzfile.h"
+#include "cleanup.h"
 
-#ifndef O_CLOEXEC
-#define O_CLOEXEC 0
-#endif
+#include "xzfile.h"
 
 #define XZ_HEADER_MAGIC     "\xfd" "7zXZ\0"
 #define XZ_HEADER_MAGIC_LEN 6
@@ -119,6 +117,10 @@ check_header_magic (struct nbdkit_next_ops *next_ops, void *nxdata)
   char buf[XZ_HEADER_MAGIC_LEN];
   int err;
 
+  if (next_ops->get_size (nxdata) < XZ_HEADER_MAGIC_LEN) {
+    nbdkit_error ("xz: file too short");
+    return false;
+  }
   if (next_ops->pread (nxdata, buf, XZ_HEADER_MAGIC_LEN, 0, 0, &err) == -1) {
     nbdkit_error ("xz: could not read header magic: error %d", err);
     return false;
@@ -360,7 +362,7 @@ xzfile_read_block (xzfile *xz,
   lzma_ret r;
   lzma_stream strm = LZMA_STREAM_INIT;
   const size_t bufsize = 1024 * 1024;
-  unsigned char *buf = NULL;
+  CLEANUP_FREE unsigned char *buf = NULL;
   char *data = NULL;
   size_t i;
 
@@ -486,8 +488,6 @@ xzfile_read_block (xzfile *xz,
   for (i = 0; filters[i].id != LZMA_VLI_UNKNOWN; ++i)
     free (filters[i].options);
 
-  free (buf);
-
   return data;
 
  err2:
@@ -497,7 +497,6 @@ xzfile_read_block (xzfile *xz,
     free (filters[i].options);
 
   free (data);
-  free (buf);
 
   return NULL;
 }
