@@ -92,11 +92,11 @@ struct nbdkit_plugin {
   int (*trim) (void *handle, uint32_t count, uint64_t offset);
   int (*zero) (void *handle, uint32_t count, uint64_t offset, int may_trim);
 #else
-  int (*_pread_old) (void *, void *, uint32_t, uint64_t);
-  int (*_pwrite_old) (void *, const void *, uint32_t, uint64_t);
-  int (*_flush_old) (void *);
-  int (*_trim_old) (void *, uint32_t, uint64_t);
-  int (*_zero_old) (void *, uint32_t, uint64_t, int);
+  int (*_pread_v1) (void *, void *, uint32_t, uint64_t);
+  int (*_pwrite_v1) (void *, const void *, uint32_t, uint64_t);
+  int (*_flush_v1) (void *);
+  int (*_trim_v1) (void *, uint32_t, uint64_t);
+  int (*_zero_v1) (void *, uint32_t, uint64_t, int);
 #endif
 
   int errno_is_preserved;
@@ -106,7 +106,7 @@ struct nbdkit_plugin {
   int (*can_zero) (void *handle);
   int (*can_fua) (void *handle);
 #if NBDKIT_API_VERSION == 1
-  int (*_unused1) (void *, void *, uint32_t, uint64_t);
+  int (*_unused1) (void *, void *, uint32_t, uint64_t, uint32_t);
   int (*_unused2) (void *, const void *, uint32_t, uint64_t, uint32_t);
   int (*_unused3) (void *, uint32_t);
   int (*_unused4) (void *, uint32_t, uint64_t, uint32_t);
@@ -128,6 +128,12 @@ struct nbdkit_plugin {
   int (*can_extents) (void *handle);
   int (*extents) (void *handle, uint32_t count, uint64_t offset, uint32_t flags,
                   struct nbdkit_extents *extents);
+  int (*can_cache) (void *handle);
+  int (*cache) (void *handle, uint32_t count, uint64_t offset, uint32_t flags);
+
+  int (*thread_model) (void);
+
+  int (*can_fast_zero) (void *handle);
 };
 
 extern void nbdkit_set_error (int err);
