@@ -46,6 +46,7 @@
 #include <stdbool.h>
 #include <stdint.h>
 #include <inttypes.h>
+#include <limits.h>
 #include <string.h>
 #include <unistd.h>
 #include <errno.h>
@@ -62,13 +63,17 @@ static const char *proxy_user = NULL;
 static char *proxy_password = NULL;
 static char *cookie = NULL;
 static bool sslverify = true;
-static long timeout = 0;
+static uint32_t timeout = 0;
 static const char *unix_socket_path = NULL;
 static long protocols = CURLPROTO_ALL;
 
 /* Use '-D curl.verbose=1' to set. */
 int curl_debug_verbose = 0;
 
+#if CURL_AT_LEAST_VERSION(7, 55, 0)
+#define HAVE_CURLINFO_CONTENT_LENGTH_DOWNLOAD_T
+#endif
+
 static void
 curl_load (void)
 {
@@ -118,8 +123,12 @@ static struct { const char *name; long bitmask; } curl_protocols[] = {
   { "rtmps", CURLPROTO_RTMPS },
   { "rtmpts", CURLPROTO_RTMPTS },
   { "gopher", CURLPROTO_GOPHER },
+#ifdef CURLPROTO_SMB
   { "smb", CURLPROTO_SMB },
+#endif
+#ifdef CURLPROTO_SMBS
   { "smbs", CURLPROTO_SMBS },
+#endif
   { NULL }
 };
 
@@ -200,10 +209,17 @@ curl_config (const char *key, const char *value)
   }
 
   else if (strcmp (key, "timeout") == 0) {
-    if (sscanf (value, "%ld", &timeout) != 1 || timeout < 0) {
-      nbdkit_error ("'timeout' must be 0 or a positive timeout in seconds");
+    if (nbdkit_parse_uint32_t ("timeout", value, &timeout) == -1)
+      return -1;
+#if LONG_MAX < UINT32_MAX
+    /* C17 5.2.4.2.1 requires that LONG_MAX is at least 2^31 - 1.
+     * However a large positive number might still exceed the limit.
+     */
+    if (timeout > LONG_MAX) {
+      nbdkit_error ("timeout is too large");
       return -1;
     }
+#endif
   }
 
   else if (strcmp (key, "unix-socket-path") == 0 ||
@@ -277,7 +293,11 @@ curl_open (int readonly)
 {
   struct curl_handle *h;
   CURLcode r;
+#ifdef HAVE_CURLINFO_CONTENT_LENGTH_DOWNLOAD_T
+  curl_off_t o;
+#else
   double d;
+#endif
 
   h = calloc (1, sizeof *h);
   if (h == NULL) {
@@ -330,8 +350,12 @@ curl_open (int readonly)
     curl_easy_setopt (h->c, CURLOPT_REDIR_PROTOCOLS, protocols);
   }
   if (timeout > 0)
-    curl_easy_setopt (h->c, CURLOPT_TIMEOUT, timeout);
+    /* NB: The cast is required here because the parameter is varargs
+     * treated as long, and not type safe.
+     */
+    curl_easy_setopt (h->c, CURLOPT_TIMEOUT, (long) timeout);
   if (!sslverify) {
+    /* See comment above. */
     curl_easy_setopt (h->c, CURLOPT_SSL_VERIFYPEER, 0L);
     curl_easy_setopt (h->c, CURLOPT_SSL_VERIFYHOST, 0L);
   }
@@ -361,6 +385,21 @@ curl_open (int readonly)
     goto err;
   }
 
+#ifdef HAVE_CURLINFO_CONTENT_LENGTH_DOWNLOAD_T
+  r = curl_easy_getinfo (h->c, CURLINFO_CONTENT_LENGTH_DOWNLOAD_T, &o);
+  if (r != CURLE_OK) {
+    display_curl_error (h, r, "could not get length of remote file [%s]", url);
+    goto err;
+  }
+
+  if (o == -1) {
+    nbdkit_error ("could not get length of remote file [%s], "
+                  "is the URL correct?", url);
+    goto err;
+  }
+
+  h->exportsize = o;
+#else
   r = curl_easy_getinfo (h->c, CURLINFO_CONTENT_LENGTH_DOWNLOAD, &d);
   if (r != CURLE_OK) {
     display_curl_error (h, r, "could not get length of remote file [%s]", url);
@@ -373,7 +412,8 @@ curl_open (int readonly)
     goto err;
   }
 
-  h->exportsize = (size_t) d;
+  h->exportsize = d;
+#endif
   nbdkit_debug ("content length: %" PRIi64, h->exportsize);
 
   if (strncasecmp (url, "http://", strlen ("http://")) == 0 ||
