@@ -44,8 +44,11 @@
 #include <gnutls/gnutls.h>
 #endif
 
+#define NBDKIT_API_VERSION 2
+
 #include <nbdkit-plugin.h>
 
+#include "cleanup.h"
 #include "sparse.h"
 
 /* If raw|base64|data parameter seen. */
@@ -157,7 +160,7 @@ read_data (const char *value)
   size_t i, len = strlen (value);
 
   for (i = 0; i < len; ++i) {
-    int64_t j;
+    int64_t j, k;
     int n;
     char c, cc[2];
 
@@ -169,11 +172,33 @@ read_data (const char *value)
       i += n;
       offset = j;
     }
+    else if (sscanf (&value[i], " %" SCNi64 "*%" SCNi64 "%n",
+                     &j, &k, &n) == 2) {
+      if (j < 0 || j > 255) {
+        nbdkit_error ("data parameter BYTE must be in the range 0..255");
+        return -1;
+      }
+      if (k < 0) {
+        nbdkit_error ("data parameter *N must be >= 0");
+        return -1;
+      }
+      i += n;
+
+      c = j;
+      while (k > 0) {
+        if (sparse_array_write (sa, &c, 1, offset) == -1)
+          return -1;
+        offset++;
+        k--;
+      }
+      if (data_size < offset)
+        data_size = offset;
+    }
     /* We need %1s for obscure reasons.  sscanf " <%n" can return 0
      * if nothing is matched, not only if the '<' is matched.
      */
     else if (sscanf (&value[i], " <%1s%n", cc, &n) == 1) {
-      char *filename;
+      CLEANUP_FREE char *filename = NULL;
       size_t flen;
 
       i += n-1;
@@ -191,11 +216,8 @@ read_data (const char *value)
       }
       i += len;
 
-      if (store_file (filename, &offset) == -1) {
-        free (filename);
+      if (store_file (filename, &offset) == -1)
         return -1;
-      }
-      free (filename);
 
       if (data_size < offset)
         data_size = offset;
@@ -311,14 +333,11 @@ data_dump_plugin (void)
 
 #define THREAD_MODEL NBDKIT_THREAD_MODEL_PARALLEL
 
-/* No meaning, just used as the address for the handle. */
-static int dh;
-
 /* Create the per-connection handle. */
 static void *
 data_open (int readonly)
 {
-  return &dh;
+  return NBDKIT_HANDLE_NOT_NEEDED;
 }
 
 /* Get the disk size. */
@@ -328,6 +347,13 @@ data_get_size (void *handle)
   return size;
 }
 
+/* Flush is a no-op, so advertise native FUA support */
+static int
+data_can_fua (void *handle)
+{
+  return NBDKIT_FUA_NATIVE;
+}
+
 /* Serves the same data over multiple connections. */
 static int
 data_can_multi_conn (void *handle)
@@ -335,45 +361,73 @@ data_can_multi_conn (void *handle)
   return 1;
 }
 
+/* Cache. */
+static int
+data_can_cache (void *handle)
+{
+  /* Everything is already in memory, returning this without
+   * implementing .cache lets nbdkit do the correct no-op.
+   */
+  return NBDKIT_CACHE_NATIVE;
+}
+
+/* Fast zero. */
+static int
+data_can_fast_zero (void *handle)
+{
+  return 1;
+}
+
 /* Read data. */
 static int
-data_pread (void *handle, void *buf, uint32_t count, uint64_t offset)
+data_pread (void *handle, void *buf, uint32_t count, uint64_t offset,
+            uint32_t flags)
 {
-  pthread_mutex_lock (&lock);
+  assert (!flags);
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
   sparse_array_read (sa, buf, count, offset);
-  pthread_mutex_unlock (&lock);
   return 0;
 }
 
 /* Write data. */
 static int
-data_pwrite (void *handle, const void *buf, uint32_t count, uint64_t offset)
+data_pwrite (void *handle, const void *buf, uint32_t count, uint64_t offset,
+             uint32_t flags)
 {
-  int r;
-
-  pthread_mutex_lock (&lock);
-  r = sparse_array_write (sa, buf, count, offset);
-  pthread_mutex_unlock (&lock);
-  return r;
+  /* Flushing, and thus FUA flag, is a no-op */
+  assert ((flags & ~NBDKIT_FLAG_FUA) == 0);
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
+  return sparse_array_write (sa, buf, count, offset);
 }
 
 /* Zero. */
 static int
-data_zero (void *handle, uint32_t count, uint64_t offset, int may_trim)
+data_zero (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
 {
-  pthread_mutex_lock (&lock);
+  /* Flushing, and thus FUA flag, is a no-op. Assume that
+   * sparse_array_zero generally beats writes, so FAST_ZERO is a no-op. */
+  assert ((flags & ~(NBDKIT_FLAG_FUA | NBDKIT_FLAG_MAY_TRIM |
+                     NBDKIT_FLAG_FAST_ZERO)) == 0);
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
   sparse_array_zero (sa, count, offset);
-  pthread_mutex_unlock (&lock);
   return 0;
 }
 
 /* Trim (same as zero). */
 static int
-data_trim (void *handle, uint32_t count, uint64_t offset)
+data_trim (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
 {
-  pthread_mutex_lock (&lock);
+  /* Flushing, and thus FUA flag, is a no-op */
+  assert ((flags & ~NBDKIT_FLAG_FUA) == 0);
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
   sparse_array_zero (sa, count, offset);
-  pthread_mutex_unlock (&lock);
+  return 0;
+}
+
+/* Nothing is persistent, so flush is trivially supported */
+static int
+data_flush (void *handle, uint32_t flags)
+{
   return 0;
 }
 
@@ -382,12 +436,8 @@ static int
 data_extents (void *handle, uint32_t count, uint64_t offset,
               uint32_t flags, struct nbdkit_extents *extents)
 {
-  int r;
-
-  pthread_mutex_lock (&lock);
-  r = sparse_array_extents (sa, count, offset, extents);
-  pthread_mutex_unlock (&lock);
-  return r;
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lock);
+  return sparse_array_extents (sa, count, offset, extents);
 }
 
 static struct nbdkit_plugin plugin = {
@@ -402,10 +452,14 @@ static struct nbdkit_plugin plugin = {
   .open              = data_open,
   .get_size          = data_get_size,
   .can_multi_conn    = data_can_multi_conn,
+  .can_fua           = data_can_fua,
+  .can_cache         = data_can_cache,
+  .can_fast_zero     = data_can_fast_zero,
   .pread             = data_pread,
   .pwrite            = data_pwrite,
   .zero              = data_zero,
   .trim              = data_trim,
+  .flush             = data_flush,
   .extents           = data_extents,
   /* In this plugin, errno is preserved properly along error return
    * paths from failed system calls.
