@@ -48,10 +48,6 @@
 
 #include <nbdkit-plugin.h>
 
-#ifndef O_CLOEXEC
-#define O_CLOEXEC 0
-#endif
-
 static char *filename = NULL;
 
 /* A debug flag which can be set on the command line using
