@@ -45,8 +45,10 @@
 
 #include <nbdkit-plugin.h>
 
+#include "cleanup.h"
+
 /* Configuration. */
-static const char *connect = NULL; /* libvirt URI */
+static const char *libvirt_uri = NULL; /* libvirt URI */
 static const char *export = NULL;  /* export device or file */
 static const char *format = NULL;  /* format parameter */
 static int trace = 0, debug = 0;
@@ -58,7 +60,7 @@ struct drive {
   const char *value;
   const char *format;
 };
-struct drive *drives = NULL;
+struct drive *all_drives = NULL;
 
 /* mount options.  (NB: list stored in reverse order)  */
 struct mount {
@@ -67,25 +69,27 @@ struct mount {
   const char *dev;
   const char *mp;
 };
-struct mount *mounts = NULL;
+struct mount *all_mounts = NULL;
 
 static int
 plugin_guestfs_config (const char *key, const char *value)
 {
   if (strcmp (key, "debug") == 0) {
-    if (sscanf (value, "%d", &debug) != 1) {
+    debug = nbdkit_parse_bool (value);
+    if (debug == -1) {
       nbdkit_error ("could not parse 'debug' option, expecting an integer");
       return -1;
     }
   }
   else if (strcmp (key, "trace") == 0) {
-    if (sscanf (value, "%d", &trace) != 1) {
+    trace = nbdkit_parse_bool (value);
+    if (trace == -1) {
       nbdkit_error ("could not parse 'trace' option, expecting an integer");
       return -1;
     }
   }
   else if (strcmp (key, "connect") == 0) {
-    connect = value;
+    libvirt_uri = value;
   }
   else if (strcmp (key, "export") == 0) {
     export = value;
@@ -107,8 +111,8 @@ plugin_guestfs_config (const char *key, const char *value)
     d->type = drv_disk;
     d->value = value;
     d->format = format;
-    d->next = drives;
-    drives = d;
+    d->next = all_drives;
+    all_drives = d;
   }
   else if (strcmp (key, "domain") == 0) {
     struct drive *d;
@@ -120,8 +124,8 @@ plugin_guestfs_config (const char *key, const char *value)
     }
     d->type = drv_domain;
     d->value = value;
-    d->next = drives;
-    drives = d;
+    d->next = all_drives;
+    all_drives = d;
   }
   else if (strcmp (key, "mount") == 0) {
     struct mount *m;
@@ -148,8 +152,8 @@ plugin_guestfs_config (const char *key, const char *value)
       m->dev = value;
       m->mp = "/";
     }
-    m->next = mounts;
-    mounts = m;
+    m->next = all_mounts;
+    all_mounts = m;
   }
   else {
     nbdkit_error ("unknown parameter '%s'", key);
@@ -167,7 +171,7 @@ plugin_guestfs_config_complete (void)
     return -1;
   }
 
-  if (drives == NULL) {
+  if (all_drives == NULL) {
     nbdkit_error ("at least one 'disk' or 'domain' parameter is required");
     return -1;
   }
@@ -191,12 +195,12 @@ plugin_guestfs_unload (void)
   struct drive *d, *d_next;
   struct mount *m, *m_next;
 
-  for (d = drives; d != NULL; d = d_next) {
+  for (d = all_drives; d != NULL; d = d_next) {
     d_next = d->next;
     free (d);
   }
 
-  for (m = mounts; m != NULL; m = m_next) {
+  for (m = all_mounts; m != NULL; m = m_next) {
     m_next = m->next;
     free (m);
   }
@@ -255,7 +259,7 @@ plugin_guestfs_open (int readonly)
   if (set_up_logging (h->g) == -1)
     goto err2;
 
-  if (add_disks (h->g, readonly, drives) == -1)
+  if (add_disks (h->g, readonly, all_drives) == -1)
     goto err2;
 
   if (guestfs_launch (h->g) == -1) {
@@ -263,7 +267,7 @@ plugin_guestfs_open (int readonly)
     goto err2;
   }
 
-  if (mount_filesystems (h->g, readonly, mounts) == -1)
+  if (mount_filesystems (h->g, readonly, all_mounts) == -1)
     goto err2;
 
   /* Exported thing. */
@@ -307,14 +311,12 @@ log_to_nbdkit (guestfs_h *g,
                const char *buf, size_t buf_len,
                const uint64_t *array, size_t array_len)
 {
-  char *sbuf;
+  CLEANUP_FREE char *sbuf = NULL;
 
   /* Note the buffer may not be \0 terminated.  Hence this. */
   sbuf = strndup (buf, buf_len);
-  if (sbuf) {
+  if (sbuf)
     nbdkit_debug ("%s", sbuf);
-    free (sbuf);
-  }
 }
 
 static int
@@ -355,9 +357,9 @@ add_disks (guestfs_h *g, int readonly, struct drive *drives)
       GUESTFS_ADD_DOMAIN_ALLOWUUID_BITMASK;
     domain_optargs.readonly = readonly;
     domain_optargs.allowuuid = 1;
-    if (connect) {
+    if (libvirt_uri) {
       domain_optargs.bitmask |= GUESTFS_ADD_DOMAIN_LIBVIRTURI_BITMASK;
-      domain_optargs.libvirturi = connect;
+      domain_optargs.libvirturi = libvirt_uri;
     }
     if (guestfs_add_domain_argv (g, drives->value, &domain_optargs) == -1) {
       GERROR (g, "domain %s", drives->value);
@@ -509,7 +511,6 @@ plugin_guestfs_pread (void *handle, void *buf, uint32_t count, uint64_t offset)
 {
   struct handle *h = handle;
   char *(*pr) (guestfs_h *, const char *, int, int64_t, size_t *);
-  char *data;
   size_t size;
 
   if (h->is_device)
@@ -518,7 +519,7 @@ plugin_guestfs_pread (void *handle, void *buf, uint32_t count, uint64_t offset)
     pr = guestfs_pread;
 
   while (count > 0) {
-    data = pr (h->g, export, count, offset, &size);
+    CLEANUP_FREE char *data = pr (h->g, export, count, offset, &size);
     if (!data) {
       GERROR (h->g, "%s: pread", export);
       errno = guestfs_last_errno (h->g) ? : EIO;
@@ -526,7 +527,6 @@ plugin_guestfs_pread (void *handle, void *buf, uint32_t count, uint64_t offset)
     }
 
     memcpy (buf, data, size);
-    free (data);
 
     buf += size;
     offset += size;
