@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2017-2018 Red Hat Inc.
+ * Copyright (C) 2017-2019 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -83,6 +83,30 @@ null_get_size (void *handle)
   return size;
 }
 
+/* Serves the same data over multiple connections. */
+static int
+null_can_multi_conn (void *handle)
+{
+  return 1;
+}
+
+/* Cache. */
+static int
+null_can_cache (void *handle)
+{
+  /* Everything is already in memory, returning this without
+   * implementing .cache lets nbdkit do the correct no-op.
+   */
+  return NBDKIT_CACHE_NATIVE;
+}
+
+/* Fast zero. */
+static int
+null_can_fast_zero (void *handle)
+{
+  return 1;
+}
+
 /* Read data. */
 static int
 null_pread (void *handle, void *buf, uint32_t count, uint64_t offset,
@@ -145,8 +169,12 @@ static struct nbdkit_plugin plugin = {
   .version           = PACKAGE_VERSION,
   .config            = null_config,
   .config_help       = null_config_help,
+  .magic_config_key  = "size",
   .open              = null_open,
   .get_size          = null_get_size,
+  .can_multi_conn    = null_can_multi_conn,
+  .can_cache         = null_can_cache,
+  .can_fast_zero     = null_can_fast_zero,
   .pread             = null_pread,
   .pwrite            = null_pwrite,
   .zero              = null_zero,
