@@ -39,6 +39,7 @@
 
 #include <caml/alloc.h>
 #include <caml/callback.h>
+#include <caml/fail.h>
 #include <caml/memory.h>
 #include <caml/mlvalues.h>
 #include <caml/printexc.h>
@@ -62,13 +63,16 @@ constructor (void)
 }
 
 /* Instead of using the NBDKIT_REGISTER_PLUGIN macro, we construct the
- * nbdkit_plugin struct and return it.
+ * nbdkit_plugin struct and return it from our own plugin_init
+ * function.
  */
 static void unload_wrapper (void);
+static void remove_roots (void);
 
 static struct nbdkit_plugin plugin = {
   ._struct_size = sizeof (plugin),
   ._api_version = NBDKIT_API_VERSION,
+  ._thread_model = NBDKIT_THREAD_MODEL_PARALLEL,
 
   /* The following field is used as a canary to detect whether the
    * OCaml code started up and called us back successfully.  If it's
@@ -80,6 +84,16 @@ static struct nbdkit_plugin plugin = {
   .unload = unload_wrapper,
 };
 
+struct nbdkit_plugin *
+plugin_init (void)
+{
+  if (plugin.name == NULL) {
+    fprintf (stderr, "error: OCaml code did not call NBDKit.register_plugin\n");
+    exit (EXIT_FAILURE);
+  }
+  return &plugin;
+}
+
 /* These globals store the OCaml functions that we actually call.
  * Also the assigned ones are roots to ensure the GC doesn't free them.
  */
@@ -112,6 +126,17 @@ static value zero_fn;
 
 static value can_multi_conn_fn;
 
+static value can_extents_fn;
+static value extents_fn;
+
+static value can_cache_fn;
+static value cache_fn;
+
+static value thread_model_fn;
+
+static value can_fast_zero_fn;
+
+/*----------------------------------------------------------------------*/
 /* Wrapper functions that translate calls from C (ie. nbdkit) to OCaml. */
 
 static void
@@ -140,36 +165,7 @@ unload_wrapper (void)
   free ((char *) plugin.description);
   free ((char *) plugin.config_help);
 
-#define REMOVE(fn) \
-  if (fn##_fn) caml_remove_generational_global_root (&fn##_fn)
-  REMOVE (load);
-  REMOVE (unload);
-
-  REMOVE (config);
-  REMOVE (config_complete);
-
-  REMOVE (open);
-  REMOVE (close);
-
-  REMOVE (get_size);
-
-  REMOVE (can_write);
-  REMOVE (can_flush);
-  REMOVE (is_rotational);
-  REMOVE (can_trim);
-
-  REMOVE (dump_plugin);
-
-  REMOVE (can_zero);
-  REMOVE (can_fua);
-
-  REMOVE (pread);
-  REMOVE (pwrite);
-  REMOVE (flush);
-  REMOVE (trim);
-  REMOVE (zero);
-
-  REMOVE (can_multi_conn);
+  remove_roots ();
 }
 
 static int
@@ -430,6 +426,12 @@ Val_flags (uint32_t flags)
     Store_field (consv, 1, rv);
     rv = consv;
   }
+  if (flags & NBDKIT_FLAG_REQ_ONE) {
+    consv = caml_alloc (2, 0);
+    Store_field (consv, 0, 2); /* 2 = Req_one */
+    Store_field (consv, 1, rv);
+    rv = consv;
+  }
 
   CAMLreturn (rv);
 }
@@ -439,19 +441,16 @@ pread_wrapper (void *h, void *buf, uint32_t count, uint64_t offset,
                uint32_t flags)
 {
   CAMLparam0 ();
-  CAMLlocal4 (rv, strv, offsetv, flagsv);
+  CAMLlocal4 (rv, countv, offsetv, flagsv);
+  mlsize_t len;
 
   caml_leave_blocking_section ();
 
-  strv = caml_alloc_string (count);
-  /* Initialize the buffer with zeroes in case the plugin does not
-   * fill it completely.
-   */
-  memset (String_val (strv), 0, count);
+  countv = caml_copy_int32 (count);
   offsetv = caml_copy_int64 (offset);
   flagsv = Val_flags (flags);
 
-  value args[] = { *(value *) h, strv, offsetv, flagsv };
+  value args[] = { *(value *) h, countv, offsetv, flagsv };
   rv = caml_callbackN_exn (pread_fn, sizeof args / sizeof args[0], args);
   if (Is_exception_result (rv)) {
     nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
@@ -459,7 +458,14 @@ pread_wrapper (void *h, void *buf, uint32_t count, uint64_t offset,
     CAMLreturnT (int, -1);
   }
 
-  memcpy (buf, String_val (strv), count);
+  len = caml_string_length (rv);
+  if (len < count) {
+    nbdkit_error ("buffer returned from pread is too small");
+    caml_enter_blocking_section ();
+    CAMLreturnT (int, -1);
+  }
+
+  memcpy (buf, String_val (rv), count);
 
   caml_enter_blocking_section ();
   CAMLreturnT (int, 0);
@@ -475,7 +481,7 @@ pwrite_wrapper (void *h, const void *buf, uint32_t count, uint64_t offset,
   caml_leave_blocking_section ();
 
   strv = caml_alloc_string (count);
-  memcpy (String_val (strv), buf, count);
+  memcpy (&Byte_u (strv, 0), buf, count);
   offsetv = caml_copy_int64 (offset);
   flagsv = Val_flags (flags);
 
@@ -576,13 +582,155 @@ can_multi_conn_wrapper (void *h)
   CAMLreturnT (int, Bool_val (rv));
 }
 
-value
-ocaml_nbdkit_set_thread_model (value modelv)
+static int
+can_extents_wrapper (void *h)
 {
-  plugin._thread_model = Int_val (modelv);
-  return Val_unit;
+  CAMLparam0 ();
+  CAMLlocal1 (rv);
+
+  caml_leave_blocking_section ();
+
+  rv = caml_callback_exn (can_extents_fn, *(value *) h);
+  if (Is_exception_result (rv)) {
+    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
+    caml_enter_blocking_section ();
+    CAMLreturnT (int, -1);
+  }
+
+  caml_enter_blocking_section ();
+  CAMLreturnT (int, Bool_val (rv));
 }
 
+static int
+extents_wrapper (void *h, uint32_t count, uint64_t offset, uint32_t flags,
+                 struct nbdkit_extents *extents)
+{
+  CAMLparam0 ();
+  CAMLlocal5 (rv, countv, offsetv, flagsv, v);
+
+  caml_leave_blocking_section ();
+
+  countv = caml_copy_int32 (count);
+  offsetv = caml_copy_int32 (offset);
+  flagsv = Val_flags (flags);
+
+  value args[] = { *(value *) h, countv, offsetv, flagsv };
+  rv = caml_callbackN_exn (extents_fn, sizeof args / sizeof args[0], args);
+  if (Is_exception_result (rv)) {
+    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
+    caml_enter_blocking_section ();
+    CAMLreturnT (int, -1);
+  }
+
+  /* Convert extents list into calls to nbdkit_add_extent. */
+  while (rv != Val_int (0)) {
+    uint64_t length;
+    uint32_t type = 0;
+
+    v = Field (rv, 0);          /* extent struct */
+    offset = Int64_val (Field (v, 0));
+    length = Int64_val (Field (v, 1));
+    if (Bool_val (Field (v, 2)))
+      type |= NBDKIT_EXTENT_HOLE;
+    if (Bool_val (Field (v, 3)))
+      type |= NBDKIT_EXTENT_ZERO;
+    if (nbdkit_add_extent (extents, offset, length, type) == -1) {
+      caml_enter_blocking_section ();
+      CAMLreturnT (int, -1);
+    }
+
+    rv = Field (rv, 1);
+  }
+
+  caml_enter_blocking_section ();
+  CAMLreturnT (int, 0);
+}
+
+static int
+can_cache_wrapper (void *h)
+{
+  CAMLparam0 ();
+  CAMLlocal1 (rv);
+
+  caml_leave_blocking_section ();
+
+  rv = caml_callback_exn (can_cache_fn, *(value *) h);
+  if (Is_exception_result (rv)) {
+    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
+    caml_enter_blocking_section ();
+    CAMLreturnT (int, -1);
+  }
+
+  caml_enter_blocking_section ();
+  CAMLreturnT (int, Int_val (rv));
+}
+
+static int
+cache_wrapper (void *h, uint32_t count, uint64_t offset, uint32_t flags)
+{
+  CAMLparam0 ();
+  CAMLlocal4 (rv, countv, offsetv, flagsv);
+
+  caml_leave_blocking_section ();
+
+  countv = caml_copy_int32 (count);
+  offsetv = caml_copy_int32 (offset);
+  flagsv = Val_flags (flags);
+
+  value args[] = { *(value *) h, countv, offsetv, flagsv };
+  rv = caml_callbackN_exn (cache_fn, sizeof args / sizeof args[0], args);
+  if (Is_exception_result (rv)) {
+    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
+    CAMLreturnT (int, -1);
+  }
+
+  caml_enter_blocking_section ();
+  CAMLreturnT (int, 0);
+}
+
+static int
+thread_model_wrapper (void)
+{
+  CAMLparam0 ();
+  CAMLlocal1 (rv);
+
+  caml_leave_blocking_section ();
+
+  rv = caml_callback_exn (thread_model_fn, Val_unit);
+  if (Is_exception_result (rv)) {
+    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
+    caml_enter_blocking_section ();
+    CAMLreturnT (int, -1);
+  }
+
+  caml_enter_blocking_section ();
+  CAMLreturnT (int, Int_val (rv));
+}
+
+static int
+can_fast_zero_wrapper (void *h)
+{
+  CAMLparam0 ();
+  CAMLlocal1 (rv);
+
+  caml_leave_blocking_section ();
+
+  rv = caml_callback_exn (can_fast_zero_fn, *(value *) h);
+  if (Is_exception_result (rv)) {
+    nbdkit_error ("%s", caml_format_exception (Extract_exception (rv)));
+    caml_enter_blocking_section ();
+    CAMLreturnT (int, -1);
+  }
+
+  caml_enter_blocking_section ();
+  CAMLreturnT (int, Bool_val (rv));
+}
+
+/*----------------------------------------------------------------------*/
+/* set_* functions called from OCaml code at load time to initialize
+ * fields in the plugin struct.
+ */
+
 value
 ocaml_nbdkit_set_name (value namev)
 {
@@ -657,6 +805,68 @@ SET(zero)
 
 SET(can_multi_conn)
 
+SET(can_extents)
+SET(extents)
+
+SET(can_cache)
+SET(cache)
+
+SET(thread_model)
+
+SET(can_fast_zero)
+
+#undef SET
+
+static void
+remove_roots (void)
+{
+#define REMOVE(fn) \
+  if (fn##_fn) caml_remove_generational_global_root (&fn##_fn)
+  REMOVE (load);
+  REMOVE (unload);
+
+  REMOVE (config);
+  REMOVE (config_complete);
+
+  REMOVE (open);
+  REMOVE (close);
+
+  REMOVE (get_size);
+
+  REMOVE (can_write);
+  REMOVE (can_flush);
+  REMOVE (is_rotational);
+  REMOVE (can_trim);
+
+  REMOVE (dump_plugin);
+
+  REMOVE (can_zero);
+  REMOVE (can_fua);
+
+  REMOVE (pread);
+  REMOVE (pwrite);
+  REMOVE (flush);
+  REMOVE (trim);
+  REMOVE (zero);
+
+  REMOVE (can_multi_conn);
+
+  REMOVE (can_extents);
+  REMOVE (extents);
+
+  REMOVE (can_cache);
+  REMOVE (cache);
+
+  REMOVE (thread_model);
+
+  REMOVE (can_fast_zero);
+
+#undef REMOVE
+}
+
+/*----------------------------------------------------------------------*/
+/* Bindings for miscellaneous nbdkit_* utility functions. */
+
 /* NB: noalloc function. */
 value
 ocaml_nbdkit_set_error (value nv)
@@ -664,12 +874,18 @@ ocaml_nbdkit_set_error (value nv)
   int err;
 
   switch (Int_val (nv)) {
+    /* Host errno values that will map to NBD protocol values */
   case 1: err = EPERM; break;
   case 2: err = EIO; break;
   case 3: err = ENOMEM; break;
   case 4: err = EINVAL; break;
   case 5: err = ENOSPC; break;
   case 6: err = ESHUTDOWN; break;
+  case 7: err = EOVERFLOW; break;
+  case 8: err = EOPNOTSUPP; break;
+    /* Other errno values that server/protocol.c treats specially */
+  case 9: err = EROFS; break;
+  case 10: err = EFBIG; break;
   default: abort ();
   }
 
@@ -678,6 +894,53 @@ ocaml_nbdkit_set_error (value nv)
   return Val_unit;
 }
 
+value
+ocaml_nbdkit_parse_size (value strv)
+{
+  CAMLparam1 (strv);
+  CAMLlocal1 (rv);
+  int64_t r;
+
+  r = nbdkit_parse_size (String_val (strv));
+  if (r == -1)
+    caml_invalid_argument ("nbdkit_parse_size");
+  rv = caml_copy_int64 (r);
+
+  CAMLreturn (rv);
+}
+
+value
+ocaml_nbdkit_parse_bool (value strv)
+{
+  CAMLparam1 (strv);
+  CAMLlocal1 (rv);
+  int r;
+
+  r = nbdkit_parse_bool (String_val (strv));
+  if (r == -1)
+    caml_invalid_argument ("nbdkit_parse_bool");
+  rv = Val_bool (r);
+
+  CAMLreturn (rv);
+}
+
+value
+ocaml_nbdkit_read_password (value strv)
+{
+  CAMLparam1 (strv);
+  CAMLlocal1 (rv);
+  char *password;
+  int r;
+
+  r = nbdkit_read_password (String_val (strv), &password);
+  if (r == -1)
+    caml_invalid_argument ("nbdkit_read_password");
+  rv = caml_copy_string (password);
+  free (password);
+
+  CAMLreturn (rv);
+}
+
 /* NB: noalloc function. */
 value
 ocaml_nbdkit_debug (value strv)
@@ -686,14 +949,3 @@ ocaml_nbdkit_debug (value strv)
 
   return Val_unit;
 }
-
-/* We can't directly use NBDKIT_REGISTER_PLUGIN(). */
-struct nbdkit_plugin *
-plugin_init (void)
-{
-  if (plugin.name == NULL) {
-    fprintf (stderr, "error: OCaml code did not call NBDKit.register_plugin\n");
-    exit (EXIT_FAILURE);
-  }
-  return &plugin;
-}
