@@ -48,6 +48,11 @@
 
 #include <nbdkit-plugin.h>
 
+#include "cleanup.h"
+
+/* Use of perl.h insists on shadowing my_perl during XS(). */
+#pragma GCC diagnostic ignored "-Wshadow"
+
 static PerlInterpreter *my_perl;
 static const char *script;
 
@@ -86,7 +91,7 @@ static int
 callback_defined (const char *perl_func_name)
 {
   SV *ret;
-  char *cmd;
+  CLEANUP_FREE char *cmd = NULL;
 
   if (asprintf (&cmd, "defined &%s", perl_func_name) == -1) {
     perror ("asprintf");
@@ -94,7 +99,6 @@ callback_defined (const char *perl_func_name)
   }
 
   ret = eval_pv (cmd, FALSE);
-  free (cmd);
 
   return SvTRUE (ret);
 }
@@ -108,7 +112,7 @@ check_perl_failure (void)
   if (SvTRUE (errsv)) {
     const char *err;
     STRLEN n;
-    char *err_copy;
+    CLEANUP_FREE char *err_copy = NULL;
 
     err = SvPV (errsv, n);
 
@@ -124,7 +128,6 @@ check_perl_failure (void)
       err_copy[n-1] = '\0';
 
     nbdkit_error ("%s", err_copy);
-    free (err_copy);
 
     return -1;
   }
@@ -505,7 +508,7 @@ perl_zero (void *handle, uint32_t count, uint64_t offset, int may_trim)
     FREETMPS;
     LEAVE;
 
-    if (last_error == EOPNOTSUPP) {
+    if (last_error == EOPNOTSUPP || last_error == ENOTSUP) {
       /* When user requests this particular error, we want to
          gracefully fall back, and to accomodate both a normal return
          and an exception. */
