@@ -30,14 +30,9 @@
  * SUCH DAMAGE.
  */
 
-/* This has to be included first, else definitions conflict with
- * glibc header files.  Python is broken.
- */
-#define PY_SSIZE_T_CLEAN 1
-#include <Python.h>
-
 #include <config.h>
 
+#include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
@@ -48,8 +43,13 @@
 #include <assert.h>
 #include <errno.h>
 
+#define PY_SSIZE_T_CLEAN 1
+#include <Python.h>
+
 #include <nbdkit-plugin.h>
 
+#include "cleanup.h"
+
 /* XXX Apparently global state is technically wrong in Python 3, see:
  *
  * https://www.python.org/dev/peps/pep-3121/
@@ -114,17 +114,9 @@ static char *
 python_to_string (PyObject *str)
 {
   if (str) {
-#ifdef HAVE_PYUNICODE_ASUTF8
     if (PyUnicode_Check (str))
       return strdup (PyUnicode_AsUTF8 (str));
-    else
-#endif
-#ifdef HAVE_PYSTRING_ASSTRING
-    if (PyString_Check (str))
-      return strdup (PyString_AsString (str));
-    else
-#endif
-    if (PyBytes_Check (str))
+    else if (PyBytes_Check (str))
       return strdup (PyBytes_AS_STRING (str));
   }
   return NULL;
@@ -135,7 +127,7 @@ static void
 print_python_error (const char *callback, PyObject *error)
 {
   PyObject *error_str;
-  char *error_cstr = NULL;
+  CLEANUP_FREE char *error_cstr = NULL;
 
   error_str = PyObject_Str (error);
   error_cstr = python_to_string (error_str);
@@ -143,7 +135,6 @@ print_python_error (const char *callback, PyObject *error)
                 script, callback,
                 error_cstr ? error_cstr : "<unknown>");
   Py_DECREF (error_str);
-  free (error_cstr);
 }
 
 /* Convert the Python traceback to a string and call nbdkit_error.
@@ -155,13 +146,9 @@ print_python_traceback (const char *callback,
 {
   PyObject *module_name, *traceback_module, *format_exception_fn, *rv,
     *traceback_str;
-  char *traceback_cstr;
+  CLEANUP_FREE char *traceback_cstr = NULL;
 
-#ifdef HAVE_PYSTRING_FROMSTRING
-  module_name = PyString_FromString ("traceback");
-#else
   module_name = PyUnicode_FromString ("traceback");
-#endif
   traceback_module = PyImport_Import (module_name);
   Py_DECREF (module_name);
 
@@ -178,7 +165,9 @@ print_python_traceback (const char *callback,
 
   rv = PyObject_CallFunctionObjArgs (format_exception_fn,
                                      type, error, traceback, NULL);
-  traceback_str = PyObject_Str (rv);
+  if (rv == NULL)
+    return -1;
+  traceback_str = PyUnicode_Join (NULL, rv);
   Py_DECREF (rv);
   traceback_cstr = python_to_string (traceback_str);
   if (traceback_cstr == NULL) {
@@ -190,7 +179,6 @@ print_python_traceback (const char *callback,
                 script, callback,
                 traceback_cstr);
   Py_DECREF (traceback_str);
-  free (traceback_cstr);
 
   /* This means we succeeded in calling nbdkit_error. */
   return 0;
@@ -221,7 +209,6 @@ check_python_failure (const char *callback)
   return 0;
 }
 
-#if PY_MAJOR_VERSION >= 3
 static struct PyModuleDef moduledef = {
   PyModuleDef_HEAD_INIT,
   "nbdkit",
@@ -233,25 +220,18 @@ static struct PyModuleDef moduledef = {
   NULL,
   NULL
 };
-#endif
 
 static PyMODINIT_FUNC
 create_nbdkit_module (void)
 {
   PyObject *m;
 
-#if PY_MAJOR_VERSION >= 3
   m = PyModule_Create (&moduledef);
-#else
-  m = Py_InitModule ("nbdkit", NbdkitMethods);
-#endif
   if (m == NULL) {
     nbdkit_error ("could not create the nbdkit API module");
     exit (EXIT_FAILURE);
   }
-#if PY_MAJOR_VERSION >= 3
   return m;
-#endif
 }
 
 static void
@@ -275,13 +255,8 @@ py_dump_plugin (void)
   PyObject *fn;
   PyObject *r;
 
-#ifdef PY_VERSION
   printf ("python_version=%s\n", PY_VERSION);
-#endif
-
-#ifdef PYTHON_ABI_VERSION
   printf ("python_pep_384_abi_version=%d\n", PYTHON_ABI_VERSION);
-#endif
 
   if (script && callback_defined ("dump_plugin", &fn)) {
     PyErr_Clear ();
@@ -295,6 +270,7 @@ py_dump_plugin (void)
 static int
 py_config (const char *key, const char *value)
 {
+  int fd;
   FILE *fp;
   PyObject *modname;
   PyObject *fn;
@@ -309,10 +285,22 @@ py_config (const char *key, const char *value)
     }
     script = value;
 
-    /* Load the Python script. */
-    fp = fopen (script, "r");
+    /* Load the Python script. Mark the file CLOEXEC, in case the act
+     * of loading the script invokes code that in turn fork()s.
+     * However, we can't rely on fopen("re"), so do it by hand.  This
+     * does not have to be atomic, because there are no threads during
+     * .config before the python interpreter is running, but it's
+     * easier to use open/fdopen than fopen/fcntl(fileno).
+     */
+    fd = open (script, O_CLOEXEC | O_RDONLY);
+    if (fd == -1) {
+      nbdkit_error ("%s: cannot open file: %m", script);
+      return -1;
+    }
+    fp = fdopen (fd, "r");
     if (!fp) {
       nbdkit_error ("%s: cannot open file: %m", script);
+      close (fd);
       return -1;
     }
 
@@ -323,11 +311,7 @@ py_config (const char *key, const char *value)
     /* Note that because closeit flag == 1, fp is now closed. */
 
     /* The script should define a module called __main__. */
-#ifdef HAVE_PYSTRING_FROMSTRING
-    modname = PyString_FromString ("__main__");
-#else
     modname = PyUnicode_FromString ("__main__");
-#endif
     module = PyImport_Import (modname);
     Py_DECREF (modname);
     if (!module) {
@@ -461,36 +445,43 @@ py_pread (void *handle, void *buf,
   PyObject *obj = handle;
   PyObject *fn;
   PyObject *r;
+  Py_buffer view = {0};
+  int ret = -1;
 
   if (!callback_defined ("pread", &fn)) {
     nbdkit_error ("%s: missing callback: %s", script, "pread");
-    return -1;
+    return ret;
   }
 
   PyErr_Clear ();
 
-  r = PyObject_CallFunction (fn, "OiL", obj, count, offset, NULL);
+  r = PyObject_CallFunction (fn, "OiL", obj, count, offset);
   Py_DECREF (fn);
   if (check_python_failure ("pread") == -1)
-    return -1;
+    return ret;
 
-  if (!PyByteArray_Check (r)) {
-    nbdkit_error ("%s: value returned from pread method is not a byte array",
+  if (PyObject_GetBuffer (r, &view, PyBUF_SIMPLE) == -1) {
+    nbdkit_error ("%s: value returned from pread does not support the "
+                  "buffer protocol",
                   script);
-    Py_DECREF (r);
-    return -1;
+    goto out;
   }
 
-  if (PyByteArray_Size (r) < count) {
-    nbdkit_error ("%s: byte array returned from pread is too small", script);
-    Py_DECREF (r);
-    return -1;
+  if (view.len < count) {
+    nbdkit_error ("%s: buffer returned from pread is too small", script);
+    goto out;
   }
 
-  memcpy (buf, PyByteArray_AsString (r), count);
+  memcpy (buf, view.buf, count);
+  ret = 0;
+
+out:
+  if (view.obj)
+    PyBuffer_Release (&view);
+
   Py_DECREF (r);
 
-  return 0;
+  return ret;
 }
 
 static int
@@ -505,8 +496,8 @@ py_pwrite (void *handle, const void *buf,
     PyErr_Clear ();
 
     r = PyObject_CallFunction (fn, "ONL", obj,
-                               PyByteArray_FromStringAndSize (buf, count),
-                               offset, NULL);
+            PyMemoryView_FromMemory ((char *)buf, count, PyBUF_READ),
+            offset);
     Py_DECREF (fn);
     if (check_python_failure ("pwrite") == -1)
       return -1;
@@ -554,7 +545,7 @@ py_trim (void *handle, uint32_t count, uint64_t offset)
   if (callback_defined ("trim", &fn)) {
     PyErr_Clear ();
 
-    r = PyObject_CallFunction (fn, "OiL", obj, count, offset, NULL);
+    r = PyObject_CallFunction (fn, "OiL", obj, count, offset);
     Py_DECREF (fn);
     if (check_python_failure ("trim") == -1)
       return -1;
@@ -573,26 +564,21 @@ py_zero (void *handle, uint32_t count, uint64_t offset, int may_trim)
 {
   PyObject *obj = handle;
   PyObject *fn;
-  PyObject *args;
   PyObject *r;
 
   if (callback_defined ("zero", &fn)) {
     PyErr_Clear ();
 
     last_error = 0;
-    args = PyTuple_New (4);
-    Py_INCREF (obj); /* decremented by Py_DECREF (args) */
-    PyTuple_SetItem (args, 0, obj);
-    PyTuple_SetItem (args, 1, PyLong_FromUnsignedLongLong (count));
-    PyTuple_SetItem (args, 2, PyLong_FromUnsignedLongLong (offset));
-    PyTuple_SetItem (args, 3, PyBool_FromLong (may_trim));
-    r = PyObject_CallObject (fn, args);
+    r = PyObject_CallFunction (fn, "OiLO",
+                               obj, count, offset,
+                               may_trim ? Py_True : Py_False);
     Py_DECREF (fn);
-    Py_DECREF (args);
-    if (last_error == EOPNOTSUPP) {
+    if (last_error == EOPNOTSUPP || last_error == ENOTSUP) {
       /* When user requests this particular error, we want to
-         gracefully fall back, and to accomodate both a normal return
-         and an exception. */
+       * gracefully fall back, and to accommodate both a normal return
+       * and an exception.
+       */
       nbdkit_debug ("zero requested falling back to pwrite");
       Py_XDECREF (r);
       PyErr_Clear ();
@@ -610,56 +596,29 @@ py_zero (void *handle, uint32_t count, uint64_t offset, int may_trim)
 }
 
 static int
-py_can_write (void *handle)
+boolean_callback (void *handle, const char *can_fn, const char *plain_fn)
 {
   PyObject *obj = handle;
   PyObject *fn;
   PyObject *r;
   int ret;
 
-  if (callback_defined ("can_write", &fn)) {
+  if (callback_defined (can_fn, &fn)) {
     PyErr_Clear ();
 
     r = PyObject_CallFunctionObjArgs (fn, obj, NULL);
     Py_DECREF (fn);
-    if (check_python_failure ("can_write") == -1)
+    if (check_python_failure (can_fn) == -1)
       return -1;
     ret = r == Py_True;
     Py_DECREF (r);
     return ret;
   }
-  /* No Python can_write callback, but there's a Python pwrite callback
-   * defined, so return 1.  (In C modules, nbdkit would do this).
+  /* No Python ‘can_fn’ (eg. ‘can_write’), but if there's a Python
+   * ‘plain_fn’ (eg. ‘pwrite’) callback defined, return 1.  (In C
+   * modules, nbdkit would do this).
    */
-  else if (callback_defined ("pwrite", NULL))
-    return 1;
-  else
-    return 0;
-}
-
-static int
-py_can_flush (void *handle)
-{
-  PyObject *obj = handle;
-  PyObject *fn;
-  PyObject *r;
-  int ret;
-
-  if (callback_defined ("can_flush", &fn)) {
-    PyErr_Clear ();
-
-    r = PyObject_CallFunctionObjArgs (fn, obj, NULL);
-    Py_DECREF (fn);
-    if (check_python_failure ("can_flush") == -1)
-      return -1;
-    ret = r == Py_True;
-    Py_DECREF (r);
-    return ret;
-  }
-  /* No Python can_flush callback, but there's a Python flush callback
-   * defined, so return 1.  (In C modules, nbdkit would do this).
-   */
-  else if (callback_defined ("flush", NULL))
+  else if (plain_fn && callback_defined (plain_fn, NULL))
     return 1;
   else
     return 0;
@@ -668,52 +627,25 @@ py_can_flush (void *handle)
 static int
 py_is_rotational (void *handle)
 {
-  PyObject *obj = handle;
-  PyObject *fn;
-  PyObject *r;
-  int ret;
+  return boolean_callback (handle, "is_rotational", NULL);
+}
 
-  if (callback_defined ("is_rotational", &fn)) {
-    PyErr_Clear ();
+static int
+py_can_write (void *handle)
+{
+  return boolean_callback (handle, "can_write", "pwrite");
+}
 
-    r = PyObject_CallFunctionObjArgs (fn, obj, NULL);
-    Py_DECREF (fn);
-    if (check_python_failure ("is_rotational") == -1)
-      return -1;
-    ret = r == Py_True;
-    Py_DECREF (r);
-    return ret;
-  }
-  else
-    return 0;
+static int
+py_can_flush (void *handle)
+{
+  return boolean_callback (handle, "can_flush", "flush");
 }
 
 static int
 py_can_trim (void *handle)
 {
-  PyObject *obj = handle;
-  PyObject *fn;
-  PyObject *r;
-  int ret;
-
-  if (callback_defined ("can_trim", &fn)) {
-    PyErr_Clear ();
-
-    r = PyObject_CallFunctionObjArgs (fn, obj, NULL);
-    Py_DECREF (fn);
-    if (check_python_failure ("can_trim") == -1)
-      return -1;
-    ret = r == Py_True;
-    Py_DECREF (r);
-    return ret;
-  }
-  /* No Python can_trim callback, but there's a Python trim callback
-   * defined, so return 1.  (In C modules, nbdkit would do this).
-   */
-  else if (callback_defined ("trim", NULL))
-    return 1;
-  else
-    return 0;
+  return boolean_callback (handle, "can_trim", "trim");
 }
 
 #define py_config_help \
@@ -738,9 +670,9 @@ static struct nbdkit_plugin plugin = {
   .close             = py_close,
 
   .get_size          = py_get_size,
+  .is_rotational     = py_is_rotational,
   .can_write         = py_can_write,
   .can_flush         = py_can_flush,
-  .is_rotational     = py_is_rotational,
   .can_trim          = py_can_trim,
 
   .pread             = py_pread,
