@@ -38,6 +38,7 @@
 #include <stdarg.h>
 #include <stdint.h>
 #include <inttypes.h>
+#include <limits.h>
 #include <string.h>
 #include <unistd.h>
 #include <fcntl.h>
@@ -59,7 +60,8 @@ static bool verify_remote_host = true;
 static const char *known_hosts = NULL;
 static const char **identity = NULL;
 static size_t nr_identities = 0;
-static long timeout = 0;
+static uint32_t timeout = 0;
+static bool compression = false;
 
 /* config can be:
  * NULL => parse options from default file
@@ -151,10 +153,23 @@ ssh_config (const char *key, const char *value)
   }
 
   else if (strcmp (key, "timeout") == 0) {
-    if (sscanf (value, "%ld", &timeout) != 1) {
-      nbdkit_error ("cannot parse timeout: %s", value);
+    if (nbdkit_parse_uint32_t ("timeout", value, &timeout) == -1)
+      return -1;
+#if LONG_MAX < UINT32_MAX
+    /* C17 5.2.4.2.1 requires that LONG_MAX is at least 2^31 - 1.
+     * However a large positive number might still exceed the limit.
+     */
+    if (timeout > LONG_MAX) {
+      nbdkit_error ("timeout is too large");
       return -1;
     }
+#endif
+  }
+  else if (strcmp (key, "compression") == 0) {
+    r = nbdkit_parse_bool (value);
+    if (r == -1)
+      return -1;
+    compression = r;
   }
 
   else {
@@ -188,7 +203,8 @@ ssh_config_complete (void)
   "known-hosts=<FILENAME>     Set location of known_hosts file.\n" \
   "identity=<FILENAME>        Prepend private key (identity) file.\n" \
   "timeout=SECS               Set SSH connection timeout.\n" \
-  "verify-remote-host=false   Ignore known_hosts."
+  "verify-remote-host=false   Ignore known_hosts.\n" \
+  "compression=true           Enable compression."
 
 /* The per-connection handle. */
 struct ssh_handle {
@@ -281,11 +297,11 @@ authenticate_pubkey (ssh_session session)
 }
 
 static int
-authenticate_password (ssh_session session, const char *password)
+authenticate_password (ssh_session session, const char *pass)
 {
   int rc;
 
-  rc = ssh_userauth_password (session, NULL, password);
+  rc = ssh_userauth_password (session, NULL, pass);
   if (rc == SSH_AUTH_ERROR)
     nbdkit_debug ("password authentication failed: %s",
                   ssh_get_error (session));
@@ -403,14 +419,24 @@ ssh_open (int readonly)
     }
   }
   if (timeout > 0) {
-    r = ssh_options_set (h->session, SSH_OPTIONS_TIMEOUT, &timeout);
+    long arg = timeout;
+    r = ssh_options_set (h->session, SSH_OPTIONS_TIMEOUT, &arg);
     if (r != SSH_OK) {
-      nbdkit_error ("failed to set timeout in libssh session: %ld: %s",
+      nbdkit_error ("failed to set timeout in libssh session: %" PRIu32 ": %s",
                     timeout, ssh_get_error (h->session));
       goto err;
     }
   }
 
+  if (compression) {
+    r = ssh_options_set (h->session, SSH_OPTIONS_COMPRESSION, "yes");
+    if (r != SSH_OK) {
+      nbdkit_error ("failed to enable compression in libssh session: %s",
+                    ssh_get_error (h->session));
+      goto err;
+    }
+  }
+
   /* Read SSH config or alternative file.  Must happen last so that
    * the hostname has been set already.
    */
