@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2013-2018 Red Hat Inc.
+ * Copyright (C) 2013-2019 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -45,6 +45,7 @@
 
 #include <nbdkit-plugin.h>
 
+#include "cleanup.h"
 #include "isaligned.h"
 #include "minmax.h"
 #include "rounding.h"
@@ -57,27 +58,14 @@ int vddk_debug_diskinfo;
 /* Enable debugging of extents code with: -D vddk.extents=1 */
 int vddk_debug_extents;
 
-/* The VDDK APIs that we call.  These globals are initialized when the
- * plugin is loaded (by vddk_load).
+/* For each VDDK API define a static global variable.  These globals
+ * are initialized when the plugin is loaded (by vddk_load).
  */
-static char *(*VixDiskLib_GetErrorText) (VixError err, const char *unused);
-static void (*VixDiskLib_FreeErrorText) (char *text);
-static VixError (*VixDiskLib_InitEx) (uint32_t major, uint32_t minor, VixDiskLibGenericLogFunc *log_function, VixDiskLibGenericLogFunc *warn_function, VixDiskLibGenericLogFunc *panic_function, const char *lib_dir, const char *config_file);
-static void (*VixDiskLib_Exit) (void);
-static VixDiskLibConnectParams *(*VixDiskLib_AllocateConnectParams) (void);
-static void (*VixDiskLib_FreeConnectParams) (VixDiskLibConnectParams *params);
-static VixError (*VixDiskLib_ConnectEx) (const VixDiskLibConnectParams *params, char read_only, const char *snapshot_ref, const char *transport_modes, VixDiskLibConnection *connection);
-static VixError (*VixDiskLib_Open) (const VixDiskLibConnection connection, const char *path, uint32_t flags, VixDiskLibHandle *handle);
-static const char *(*VixDiskLib_GetTransportMode) (VixDiskLibHandle handle);
-static VixError (*VixDiskLib_Close) (VixDiskLibHandle handle);
-static VixError (*VixDiskLib_Disconnect) (VixDiskLibConnection connection);
-static VixError (*VixDiskLib_GetInfo) (VixDiskLibHandle handle, VixDiskLibInfo **info);
-static void (*VixDiskLib_FreeInfo) (VixDiskLibInfo *info);
-static VixError (*VixDiskLib_Read) (VixDiskLibHandle handle, uint64_t start_sector, uint64_t nr_sectors, unsigned char *buf);
-static VixError (*VixDiskLib_Write) (VixDiskLibHandle handle, uint64_t start_sector, uint64_t nr_sectors, const unsigned char *buf);
-static VixError (*VixDiskLib_Flush) (VixDiskLibHandle handle);
-static VixError (*VixDiskLib_QueryAllocatedBlocks) (VixDiskLibHandle diskHandle, uint64_t start_sector, uint64_t nr_sectors, uint64_t chunk_size, VixDiskLibBlockList **block_list);
-static VixError (*VixDiskLib_FreeBlockList) (VixDiskLibBlockList *block_list);
+#define STUB(fn,ret,args) static ret (*fn) args
+#define OPTIONAL_STUB(fn,ret,args) static ret (*fn) args
+#include "vddk-stubs.h"
+#undef STUB
+#undef OPTIONAL_STUB
 
 /* Parameters passed to InitEx. */
 #define VDDK_MAJOR 5
@@ -90,9 +78,9 @@ static char *config = NULL;                /* config */
 static const char *cookie = NULL;          /* cookie */
 static const char *filename = NULL;        /* file */
 static char *libdir = NULL;                /* libdir */
-static int nfc_host_port = 0;              /* nfchostport */
+static uint16_t nfc_host_port = 0;         /* nfchostport */
 static char *password = NULL;              /* password */
-static int port = 0;                       /* port */
+static uint16_t port = 0;                  /* port */
 static const char *server_name = NULL;     /* server */
 static bool single_link = false;           /* single-link */
 static const char *snapshot_moref = NULL;  /* snapshot */
@@ -127,7 +115,7 @@ trim (char *str)
 static void
 debug_function (const char *fs, va_list args)
 {
-  char *str;
+  CLEANUP_FREE char *str = NULL;
 
   if (vasprintf (&str, fs, args) == -1) {
     nbdkit_debug ("lost debug message: %s", fs);
@@ -137,14 +125,13 @@ debug_function (const char *fs, va_list args)
   trim (str);
 
   nbdkit_debug ("%s", str);
-  free (str);
 }
 
 /* Turn error messages from the library into nbdkit_error. */
 static void
 error_function (const char *fs, va_list args)
 {
-  char *str;
+  CLEANUP_FREE char *str = NULL;
 
   if (vasprintf (&str, fs, args) == -1) {
     nbdkit_error ("lost error message: %s", fs);
@@ -154,7 +141,6 @@ error_function (const char *fs, va_list args)
   trim (str);
 
   nbdkit_error ("%s", str);
-  free (str);
 }
 
 /* Load and unload the plugin. */
@@ -167,7 +153,7 @@ vddk_load (void)
     "libvixDiskLib.so.5",
   };
   size_t i;
-  char *orig_error = NULL;
+  CLEANUP_FREE char *orig_error = NULL;
 
   /* Load the library. */
   for (i = 0; i < sizeof sonames / sizeof sonames[0]; ++i) {
@@ -186,49 +172,23 @@ vddk_load (void)
                   "set $LD_LIBRARY_PATH or edit /etc/ld.so.conf.\n\n"
                   "See the nbdkit-vddk-plugin(1) man page for details.",
                   orig_error ? : "(unknown error)", sonames[0]);
-    free (orig_error);
     exit (EXIT_FAILURE);
   }
-  free (orig_error);
 
-  /* Non-optional symbols.  I have checked that all these exist in at
-   * least VDDK 5.1.1 (2011) which is the earliest version of VDDK
-   * that we support.
-   */
-#define LOAD(sym)                                                 \
+  /* Load symbols. */
+#define STUB(fn,ret,args)                                         \
   do {                                                            \
-    sym = dlsym (dl, #sym);                                       \
-    if (sym == NULL) {                                            \
+    fn = dlsym (dl, #fn);                                         \
+    if (fn == NULL) {                                             \
       nbdkit_error ("required VDDK symbol \"%s\" is missing: %s", \
-                    #sym, dlerror ());                            \
+                    #fn, dlerror ());                             \
       exit (EXIT_FAILURE);                                        \
     }                                                             \
   } while (0)
-  LOAD (VixDiskLib_GetErrorText);
-  LOAD (VixDiskLib_FreeErrorText);
-  LOAD (VixDiskLib_InitEx);
-  LOAD (VixDiskLib_Exit);
-  LOAD (VixDiskLib_ConnectEx);
-  LOAD (VixDiskLib_Open);
-  LOAD (VixDiskLib_GetTransportMode);
-  LOAD (VixDiskLib_Close);
-  LOAD (VixDiskLib_Disconnect);
-  LOAD (VixDiskLib_GetInfo);
-  LOAD (VixDiskLib_FreeInfo);
-  LOAD (VixDiskLib_Read);
-  LOAD (VixDiskLib_Write);
-  LOAD (VixDiskLib_FreeConnectParams);
-#undef LOAD
-
-#define LOAD_OPTIONAL(sym) sym = dlsym (dl, #sym)
-  /* Added in VDDK 6.0, this will be NULL in earlier versions. */
-  LOAD_OPTIONAL (VixDiskLib_Flush);
-
-  /* Added in VDDK 6.7, these will be NULL for earlier versions: */
-  LOAD_OPTIONAL (VixDiskLib_QueryAllocatedBlocks);
-  LOAD_OPTIONAL (VixDiskLib_FreeBlockList);
-  LOAD_OPTIONAL (VixDiskLib_AllocateConnectParams);
-#undef LOAD_OPTIONAL
+#define OPTIONAL_STUB(fn,ret,args) fn = dlsym (dl, #fn)
+#include "vddk-stubs.h"
+#undef STUB
+#undef OPTIONAL_STUB
 }
 
 static void
@@ -274,10 +234,8 @@ vddk_config (const char *key, const char *value)
       return -1;
   }
   else if (strcmp (key, "nfchostport") == 0) {
-    if (sscanf (value, "%d", &nfc_host_port) != 1) {
-      nbdkit_error ("cannot parse nfchostport: %s", value);
+    if (nbdkit_parse_uint16_t ("nfchostport", value, &nfc_host_port) == -1)
       return -1;
-    }
   }
   else if (strcmp (key, "password") == 0) {
     free (password);
@@ -285,10 +243,8 @@ vddk_config (const char *key, const char *value)
       return -1;
   }
   else if (strcmp (key, "port") == 0) {
-    if (sscanf (value, "%d", &port) != 1) {
-      nbdkit_error ("cannot parse port: %s", value);
+    if (nbdkit_parse_uint16_t ("port", value, &port) == -1)
       return -1;
-    }
   }
   else if (strcmp (key, "server") == 0) {
     server_name = value;
@@ -411,12 +367,11 @@ vddk_dump_plugin (void)
    * contain the version number.
    */
   Dl_info info;
-  char *p;
+  CLEANUP_FREE char *p = NULL;
   if (dladdr (VixDiskLib_InitEx, &info) != 0 &&
       info.dli_fname != NULL &&
       (p = nbdkit_realpath (info.dli_fname)) != NULL) {
     printf ("vddk_dll=%s\n", p);
-    free (p);
   }
 #endif
 }
@@ -560,11 +515,11 @@ vddk_close (void *handle)
 {
   struct vddk_handle *h = handle;
 
-  free_connect_params (h->params);
   DEBUG_CALL ("VixDiskLib_Close", "handle");
   VixDiskLib_Close (h->handle);
   DEBUG_CALL ("VixDiskLib_Disconnect", "connection");
   VixDiskLib_Disconnect (h->connection);
+  free_connect_params (h->params);
   free (h);
 }
 
