@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2017-2018 Red Hat Inc.
+ * Copyright (C) 2017-2019 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -41,21 +41,11 @@
 #define NBDKIT_API_VERSION 2
 #include <nbdkit-plugin.h>
 
-static int
-zero_config (const char *key, const char *value)
-{
-  nbdkit_error ("unknown parameter '%s'", key);
-  return -1;
-}
-
-/* No meaning, just used as the address for the handle. */
-static int zero;
-
 /* Create the per-connection handle. */
 static void *
 zero_open (int readonly)
 {
-  return &zero;
+  return NBDKIT_HANDLE_NOT_NEEDED;
 }
 
 #define THREAD_MODEL NBDKIT_THREAD_MODEL_PARALLEL
@@ -67,6 +57,23 @@ zero_get_size (void *handle)
   return 0;
 }
 
+/* Serves the same data over multiple connections. */
+static int
+zero_can_multi_conn (void *handle)
+{
+  return 1;
+}
+
+/* Cache. */
+static int
+zero_can_cache (void *handle)
+{
+  /* Everything is already in memory, returning this without
+   * implementing .cache lets nbdkit do the correct no-op.
+   */
+  return NBDKIT_CACHE_NATIVE;
+}
+
 /* Ideally the read plugin would be optional. */
 static int
 zero_pread (void *handle, void *buf, uint32_t count, uint64_t offset,
@@ -79,9 +86,10 @@ zero_pread (void *handle, void *buf, uint32_t count, uint64_t offset,
 static struct nbdkit_plugin plugin = {
   .name              = "zero",
   .version           = PACKAGE_VERSION,
-  .config            = zero_config,
   .open              = zero_open,
   .get_size          = zero_get_size,
+  .can_multi_conn    = zero_can_multi_conn,
+  .can_cache         = zero_can_cache,
   .pread             = zero_pread,
   /* In this plugin, errno is preserved properly along error return
    * paths from failed system calls.
