@@ -38,8 +38,12 @@
 #include <inttypes.h>
 #include <string.h>
 #include <unistd.h>
+#include <sys/socket.h>
+#include <fcntl.h>
+#include <assert.h>
 
 #include "internal.h"
+#include "utils.h"
 
 /* Default number of parallel requests. */
 #define DEFAULT_PARALLEL_REQUESTS 16
@@ -50,40 +54,17 @@ static void free_connection (struct connection *conn);
 
 /* Don't call these raw socket functions directly.  Use conn->recv etc. */
 static int raw_recv (struct connection *, void *buf, size_t len);
-static int raw_send (struct connection *, const void *buf, size_t len);
+static int raw_send_socket (struct connection *, const void *buf, size_t len,
+                            int flags);
+static int raw_send_other (struct connection *, const void *buf, size_t len,
+                           int flags);
 static void raw_close (struct connection *);
 
-int
-connection_set_handle (struct connection *conn, size_t i, void *handle)
-{
-  size_t j;
-
-  if (i < conn->nr_handles)
-    conn->handles[i] = handle;
-  else {
-    j = conn->nr_handles;
-    conn->nr_handles = i+1;
-    conn->handles = realloc (conn->handles,
-                             conn->nr_handles * sizeof (void *));
-    if (conn->handles == NULL) {
-      perror ("realloc");
-      conn->nr_handles = 0;
-      return -1;
-    }
-    for (; j < i; ++j)
-      conn->handles[j] = NULL;
-    conn->handles[i] = handle;
-  }
-  return 0;
-}
-
 void *
 connection_get_handle (struct connection *conn, size_t i)
 {
-  if (i < conn->nr_handles)
-    return conn->handles[i];
-  else
-    return NULL;
+  assert (i < conn->nr_handles);
+  return conn->handles[i].handle;
 }
 
 int
@@ -91,11 +72,13 @@ connection_get_status (struct connection *conn)
 {
   int r;
 
-  if (conn->nworkers)
-    pthread_mutex_lock (&conn->status_lock);
+  if (conn->nworkers &&
+      pthread_mutex_lock (&conn->status_lock))
+    abort ();
   r = conn->status;
-  if (conn->nworkers)
-    pthread_mutex_unlock (&conn->status_lock);
+  if (conn->nworkers &&
+      pthread_mutex_unlock (&conn->status_lock))
+    abort ();
   return r;
 }
 
@@ -105,12 +88,22 @@ connection_get_status (struct connection *conn)
 int
 connection_set_status (struct connection *conn, int value)
 {
-  if (conn->nworkers)
-    pthread_mutex_lock (&conn->status_lock);
-  if (value < conn->status)
+  if (conn->nworkers &&
+      pthread_mutex_lock (&conn->status_lock))
+    abort ();
+  if (value < conn->status) {
+    if (conn->nworkers && conn->status > 0) {
+      char c = 0;
+
+      assert (conn->status_pipe[1] >= 0);
+      if (write (conn->status_pipe[1], &c, 1) != 1 && errno != EAGAIN)
+        debug ("failed to notify pipe-to-self: %m");
+    }
     conn->status = value;
-  if (conn->nworkers)
-    pthread_mutex_unlock (&conn->status_lock);
+  }
+  if (conn->nworkers &&
+      pthread_mutex_unlock (&conn->status_lock))
+    abort ();
   return value;
 }
 
@@ -129,6 +122,7 @@ connection_worker (void *data)
   debug ("starting worker thread %s", name);
   threadlocal_new_server_thread ();
   threadlocal_set_name (name);
+  threadlocal_set_conn (conn);
   free (worker);
 
   while (!quit && connection_get_status (conn) > 0)
@@ -138,15 +132,17 @@ connection_worker (void *data)
   return NULL;
 }
 
-static int
-_handle_single_connection (int sockin, int sockout)
+void
+handle_single_connection (int sockin, int sockout)
 {
   const char *plugin_name;
-  int ret = -1, r;
+  int r;
   struct connection *conn;
   int nworkers = threads ? threads : DEFAULT_PARALLEL_REQUESTS;
   pthread_t *workers = NULL;
 
+  lock_connection ();
+
   if (backend->thread_model (backend) < NBDKIT_THREAD_MODEL_PARALLEL ||
       nworkers == 1)
     nworkers = 0;
@@ -154,12 +150,6 @@ _handle_single_connection (int sockin, int sockout)
   if (!conn)
     goto done;
 
-  lock_request (conn);
-  r = backend->open (backend, conn, readonly);
-  unlock_request (conn);
-  if (r == -1)
-    goto done;
-
   /* NB: because of an asynchronous exit backend can be set to NULL at
    * just about any time.
    */
@@ -169,17 +159,11 @@ _handle_single_connection (int sockin, int sockout)
     plugin_name = "(unknown)";
   threadlocal_set_name (plugin_name);
 
-  /* Prepare (for filters), called just after open. */
-  lock_request (conn);
-  if (backend)
-    r = backend->prepare (backend, conn);
-  else
-    r = 0;
-  unlock_request (conn);
-  if (r == -1)
-    goto done;
-
-  /* Handshake. */
+  /* NBD handshake.
+   *
+   * Note that this calls the backend .open callback when it is safe
+   * to do so (eg. after TLS authentication).
+   */
   if (protocol_handshake (conn) == -1)
     goto done;
 
@@ -194,7 +178,7 @@ _handle_single_connection (int sockin, int sockout)
     debug ("handshake complete, processing requests with %d threads",
            nworkers);
     workers = calloc (nworkers, sizeof *workers);
-    if (!workers) {
+    if (unlikely (!workers)) {
       perror ("malloc");
       goto done;
     }
@@ -203,12 +187,13 @@ _handle_single_connection (int sockin, int sockout)
       struct worker_data *worker = malloc (sizeof *worker);
       int err;
 
-      if (!worker) {
+      if (unlikely (!worker)) {
         perror ("malloc");
         connection_set_status (conn, -1);
         goto wait;
       }
-      if (asprintf (&worker->name, "%s.%d", plugin_name, nworkers) < 0) {
+      if (unlikely (asprintf (&worker->name, "%s.%d", plugin_name, nworkers)
+                    < 0)) {
         perror ("asprintf");
         connection_set_status (conn, -1);
         free (worker);
@@ -217,7 +202,7 @@ _handle_single_connection (int sockin, int sockout)
       worker->conn = conn;
       err = pthread_create (&workers[nworkers], NULL, connection_worker,
                             worker);
-      if (err) {
+      if (unlikely (err)) {
         errno = err;
         perror ("pthread_create");
         connection_set_status (conn, -1);
@@ -234,36 +219,23 @@ _handle_single_connection (int sockin, int sockout)
 
   /* Finalize (for filters), called just before close. */
   lock_request (conn);
-  if (backend)
-    r = backend->finalize (backend, conn);
-  else
-    r = 0;
+  r = backend_finalize (backend, conn);
   unlock_request (conn);
   if (r == -1)
     goto done;
 
-  ret = connection_get_status (conn);
  done:
   free_connection (conn);
-  return ret;
-}
-
-int
-handle_single_connection (int sockin, int sockout)
-{
-  int r;
-
-  lock_connection ();
-  r = _handle_single_connection (sockin, sockout);
   unlock_connection ();
-
-  return r;
 }
 
 static struct connection *
 new_connection (int sockin, int sockout, int nworkers)
 {
   struct connection *conn;
+  int opt;
+  socklen_t optlen = sizeof opt;
+  struct backend *b;
 
   conn = calloc (1, sizeof *conn);
   if (conn == NULL) {
@@ -271,8 +243,57 @@ new_connection (int sockin, int sockout, int nworkers)
     return NULL;
   }
 
+  conn->status_pipe[0] = conn->status_pipe[1] = -1;
+
+  conn->handles = calloc (backend->i + 1, sizeof *conn->handles);
+  if (conn->handles == NULL) {
+    perror ("malloc");
+    goto error;
+  }
+  conn->nr_handles = backend->i + 1;
+  for_each_backend (b)
+    reset_b_conn_handle (&conn->handles[b->i]);
+
   conn->status = 1;
   conn->nworkers = nworkers;
+  if (nworkers) {
+#ifdef HAVE_PIPE2
+    if (pipe2 (conn->status_pipe, O_NONBLOCK | O_CLOEXEC)) {
+      perror ("pipe2");
+      goto error;
+    }
+#else
+    /* If we were fully parallel, then this function could be
+     * accepting connections in one thread while another thread could
+     * be in a plugin trying to fork.  But plugins.c forced
+     * thread_model to serialize_all_requests when it detects a lack
+     * of atomic CLOEXEC, at which point, we can use a mutex to ensure
+     * we aren't accepting until the plugin is not running, making
+     * non-atomicity okay.
+     */
+    assert (backend->thread_model (backend) <=
+            NBDKIT_THREAD_MODEL_SERIALIZE_ALL_REQUESTS);
+    lock_request (NULL);
+    if (pipe (conn->status_pipe)) {
+      perror ("pipe");
+      unlock_request (NULL);
+      goto error;
+    }
+    if (set_nonblock (set_cloexec (conn->status_pipe[0])) == -1) {
+      perror ("fcntl");
+      close (conn->status_pipe[1]);
+      unlock_request (NULL);
+      goto error;
+    }
+    if (set_nonblock (set_cloexec (conn->status_pipe[1])) == -1) {
+      perror ("fcntl");
+      close (conn->status_pipe[0]);
+      unlock_request (NULL);
+      goto error;
+    }
+    unlock_request (NULL);
+#endif
+  }
   conn->sockin = sockin;
   conn->sockout = sockout;
   pthread_mutex_init (&conn->request_lock, NULL);
@@ -281,10 +302,24 @@ new_connection (int sockin, int sockout, int nworkers)
   pthread_mutex_init (&conn->status_lock, NULL);
 
   conn->recv = raw_recv;
-  conn->send = raw_send;
+  if (getsockopt (sockout, SOL_SOCKET, SO_TYPE, &opt, &optlen) == 0)
+    conn->send = raw_send_socket;
+  else
+    conn->send = raw_send_other;
   conn->close = raw_close;
 
+  threadlocal_set_conn (conn);
+
   return conn;
+
+ error:
+  if (conn->status_pipe[0] >= 0)
+    close (conn->status_pipe[0]);
+  if (conn->status_pipe[1] >= 0)
+    close (conn->status_pipe[1]);
+  free (conn->handles);
+  free (conn);
+  return NULL;
 }
 
 static void
@@ -293,18 +328,34 @@ free_connection (struct connection *conn)
   if (!conn)
     return;
 
+  threadlocal_set_conn (NULL);
   conn->close (conn);
+  if (listen_stdin) {
+    int fd;
+
+    /* Restore something to stdin/out so the rest of our code can
+     * continue to assume that all new fds will be above stderr.
+     * Swap directions to get EBADF on improper use of stdin/out.
+     */
+    fd = open ("/dev/null", O_WRONLY | O_CLOEXEC);
+    assert (fd == 0);
+    fd = open ("/dev/null", O_RDONLY | O_CLOEXEC);
+    assert (fd == 1);
+  }
 
   /* Don't call the plugin again if quit has been set because the main
    * thread will be in the process of unloading it.  The plugin.unload
    * callback should always be called.
    */
   if (!quit) {
-    if (conn->nr_handles > 0 && conn->handles[0]) {
-      lock_request (conn);
-      backend->close (backend, conn);
-      unlock_request (conn);
-    }
+    lock_request (conn);
+    backend_close (backend, conn);
+    unlock_request (conn);
+  }
+
+  if (conn->status_pipe[0] >= 0) {
+    close (conn->status_pipe[0]);
+    close (conn->status_pipe[1]);
   }
 
   pthread_mutex_destroy (&conn->request_lock);
@@ -316,11 +367,43 @@ free_connection (struct connection *conn)
   free (conn);
 }
 
-/* Write buffer to conn->sockout and either succeed completely
- * (returns 0) or fail (returns -1).
+/* Write buffer to conn->sockout with send() and either succeed completely
+ * (returns 0) or fail (returns -1). flags may include SEND_MORE as a hint
+ * that this send will be followed by related data.
  */
 static int
-raw_send (struct connection *conn, const void *vbuf, size_t len)
+raw_send_socket (struct connection *conn, const void *vbuf, size_t len,
+                 int flags)
+{
+  int sock = conn->sockout;
+  const char *buf = vbuf;
+  ssize_t r;
+  int f = 0;
+
+#ifdef MSG_MORE
+  if (flags & SEND_MORE)
+    f |= MSG_MORE;
+#endif
+  while (len > 0) {
+    r = send (sock, buf, len, f);
+    if (r == -1) {
+      if (errno == EINTR || errno == EAGAIN)
+        continue;
+      return -1;
+    }
+    buf += r;
+    len -= r;
+  }
+
+  return 0;
+}
+
+/* Write buffer to conn->sockout with write() and either succeed completely
+ * (returns 0) or fail (returns -1). flags is ignored.
+ */
+static int
+raw_send_other (struct connection *conn, const void *vbuf, size_t len,
+                int flags)
 {
   int sock = conn->sockout;
   const char *buf = vbuf;
