@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2013-2018 Red Hat Inc.
+ * Copyright (C) 2013-2019 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -46,6 +46,8 @@ enum {
   FILTER_OPTION,
   LOG_OPTION,
   LONG_OPTIONS_OPTION,
+  MASK_HANDSHAKE_OPTION,
+  NO_SR_OPTION,
   RUN_OPTION,
   SELINUX_LABEL_OPTION,
   SHORT_OPTIONS_OPTION,
@@ -53,6 +55,7 @@ enum {
   TLS_CERTIFICATES_OPTION,
   TLS_PSK_OPTION,
   TLS_VERIFY_PEER_OPTION,
+  VSOCK_OPTION,
 };
 
 static const char *short_options = "D:e:fg:i:nop:P:rst:u:U:vV";
@@ -73,8 +76,10 @@ static const struct option long_options[] = {
   { "ipaddr",           required_argument, NULL, 'i' },
   { "log",              required_argument, NULL, LOG_OPTION },
   { "long-options",     no_argument,       NULL, LONG_OPTIONS_OPTION },
+  { "mask-handshake",   required_argument, NULL, MASK_HANDSHAKE_OPTION },
   { "new-style",        no_argument,       NULL, 'n' },
   { "newstyle",         no_argument,       NULL, 'n' },
+  { "no-sr",            no_argument,       NULL, NO_SR_OPTION },
   { "old-style",        no_argument,       NULL, 'o' },
   { "oldstyle",         no_argument,       NULL, 'o' },
   { "pid-file",         required_argument, NULL, 'P' },
@@ -96,6 +101,7 @@ static const struct option long_options[] = {
   { "user",             required_argument, NULL, 'u' },
   { "verbose",          no_argument,       NULL, 'v' },
   { "version",          no_argument,       NULL, 'V' },
+  { "vsock",            no_argument,       NULL, VSOCK_OPTION },
   { NULL },
 };
 
