@@ -41,20 +41,18 @@
 
 #include "internal.h"
 #include "byte-swapping.h"
-#include "protocol.h"
+#include "nbd-protocol.h"
+#include "protostrings.h"
 
 /* Maximum number of client options we allow before giving up. */
 #define MAX_NR_OPTIONS 32
 
-/* Maximum length of any option data (bytes). */
-#define MAX_OPTION_LENGTH 4096
-
 /* Receive newstyle options. */
 static int
 send_newstyle_option_reply (struct connection *conn,
                             uint32_t option, uint32_t reply)
 {
-  struct fixed_new_option_reply fixed_new_option_reply;
+  struct nbd_fixed_new_option_reply fixed_new_option_reply;
 
   fixed_new_option_reply.magic = htobe64 (NBD_REP_MAGIC);
   fixed_new_option_reply.option = htobe32 (option);
@@ -63,7 +61,7 @@ send_newstyle_option_reply (struct connection *conn,
 
   if (conn->send (conn,
                   &fixed_new_option_reply,
-                  sizeof fixed_new_option_reply) == -1) {
+                  sizeof fixed_new_option_reply, 0) == -1) {
     /* The protocol document says that the client is allowed to simply
      * drop the connection after sending NBD_OPT_ABORT, or may read
      * the reply.
@@ -80,10 +78,9 @@ send_newstyle_option_reply (struct connection *conn,
 
 static int
 send_newstyle_option_reply_exportname (struct connection *conn,
-                                       uint32_t option, uint32_t reply,
-                                       const char *exportname)
+                                       uint32_t option, uint32_t reply)
 {
-  struct fixed_new_option_reply fixed_new_option_reply;
+  struct nbd_fixed_new_option_reply fixed_new_option_reply;
   size_t name_len = strlen (exportname);
   uint32_t len;
 
@@ -94,18 +91,18 @@ send_newstyle_option_reply_exportname (struct connection *conn,
 
   if (conn->send (conn,
                   &fixed_new_option_reply,
-                  sizeof fixed_new_option_reply) == -1) {
+                  sizeof fixed_new_option_reply, SEND_MORE) == -1) {
     nbdkit_error ("write: %s: %m", name_of_nbd_opt (option));
     return -1;
   }
 
   len = htobe32 (name_len);
-  if (conn->send (conn, &len, sizeof len) == -1) {
+  if (conn->send (conn, &len, sizeof len, SEND_MORE) == -1) {
     nbdkit_error ("write: %s: %s: %m",
                   name_of_nbd_opt (option), "sending length");
     return -1;
   }
-  if (conn->send (conn, exportname, name_len) == -1) {
+  if (conn->send (conn, exportname, name_len, 0) == -1) {
     nbdkit_error ("write: %s: %s: %m",
                   name_of_nbd_opt (option), "sending export name");
     return -1;
@@ -117,23 +114,23 @@ send_newstyle_option_reply_exportname (struct connection *conn,
 static int
 send_newstyle_option_reply_info_export (struct connection *conn,
                                         uint32_t option, uint32_t reply,
-                                        uint16_t info)
+                                        uint16_t info, uint64_t exportsize)
 {
-  struct fixed_new_option_reply fixed_new_option_reply;
-  struct fixed_new_option_reply_info_export export;
+  struct nbd_fixed_new_option_reply fixed_new_option_reply;
+  struct nbd_fixed_new_option_reply_info_export export;
 
   fixed_new_option_reply.magic = htobe64 (NBD_REP_MAGIC);
   fixed_new_option_reply.option = htobe32 (option);
   fixed_new_option_reply.reply = htobe32 (reply);
   fixed_new_option_reply.replylen = htobe32 (sizeof export);
   export.info = htobe16 (info);
-  export.exportsize = htobe64 (conn->exportsize);
+  export.exportsize = htobe64 (exportsize);
   export.eflags = htobe16 (conn->eflags);
 
   if (conn->send (conn,
                   &fixed_new_option_reply,
-                  sizeof fixed_new_option_reply) == -1 ||
-      conn->send (conn, &export, sizeof export) == -1) {
+                  sizeof fixed_new_option_reply, SEND_MORE) == -1 ||
+      conn->send (conn, &export, sizeof export, 0) == -1) {
     nbdkit_error ("write: %s: %m", name_of_nbd_opt (option));
     return -1;
   }
@@ -147,8 +144,8 @@ send_newstyle_option_reply_meta_context (struct connection *conn,
                                          uint32_t context_id,
                                          const char *name)
 {
-  struct fixed_new_option_reply fixed_new_option_reply;
-  struct fixed_new_option_reply_meta_context context;
+  struct nbd_fixed_new_option_reply fixed_new_option_reply;
+  struct nbd_fixed_new_option_reply_meta_context context;
   const size_t namelen = strlen (name);
 
   debug ("newstyle negotiation: %s: replying with %s id %d",
@@ -161,9 +158,9 @@ send_newstyle_option_reply_meta_context (struct connection *conn,
 
   if (conn->send (conn,
                   &fixed_new_option_reply,
-                  sizeof fixed_new_option_reply) == -1 ||
-      conn->send (conn, &context, sizeof context) == -1 ||
-      conn->send (conn, name, namelen) == -1) {
+                  sizeof fixed_new_option_reply, SEND_MORE) == -1 ||
+      conn->send (conn, &context, sizeof context, SEND_MORE) == -1 ||
+      conn->send (conn, name, namelen, 0) == -1) {
     nbdkit_error ("write: %s: %m", name_of_nbd_opt (option));
     return -1;
   }
@@ -198,52 +195,88 @@ conn_recv_full (struct connection *conn, void *buf, size_t len,
 
 /* Sub-function of negotiate_handshake_newstyle_options below.  It
  * must be called on all non-error paths out of the options for-loop
- * in that function.
+ * in that function, and must not cause any wire traffic.
  */
 static int
-finish_newstyle_options (struct connection *conn)
+finish_newstyle_options (struct connection *conn, uint64_t *exportsize)
 {
-  int64_t r;
-
-  r = backend->get_size (backend, conn);
-  if (r == -1)
-    return -1;
-  if (r < 0) {
-    nbdkit_error (".get_size function returned invalid value "
-                  "(%" PRIi64 ")", r);
-    return -1;
-  }
-  conn->exportsize = (uint64_t) r;
-
-  if (protocol_compute_eflags (conn, &conn->eflags) < 0)
+  if (protocol_common_open (conn, exportsize, &conn->eflags) == -1)
     return -1;
 
   debug ("newstyle negotiation: flags: export 0x%x", conn->eflags);
   return 0;
 }
 
+/* Check that the string sent as part of @option, beginning at @buf,
+ * and with a client-reported length of @len, fits within @maxlen
+ * bytes and is well-formed.  If not, report an error mentioning
+ * @name.
+ */
+static int
+check_string (uint32_t option, char *buf, uint32_t len, uint32_t maxlen,
+              const char *name)
+{
+  if (len > NBD_MAX_STRING || len > maxlen) {
+    nbdkit_error ("%s: %s too long", name_of_nbd_opt (option), name);
+    return -1;
+  }
+  if (strnlen (buf, len) != len) {
+    nbdkit_error ("%s: %s may not include NUL bytes",
+                  name_of_nbd_opt (option), name);
+    return -1;
+  }
+  /* TODO: Check for valid UTF-8? */
+  return 0;
+}
+
+/* Sub-function of negotiate_handshake_newstyle_options, to grab and
+ * validate an export name.
+ */
+static int
+check_export_name (struct connection *conn, uint32_t option, char *buf,
+                   uint32_t exportnamelen, uint32_t maxlen, bool save)
+{
+  if (check_string (option, buf, exportnamelen, maxlen, "export name") == -1)
+    return -1;
+
+  assert (exportnamelen < sizeof conn->exportname);
+  if (save) {
+    if (exportnamelen != conn->exportnamelen ||
+        memcmp (conn->exportname, buf, exportnamelen) != 0)
+      conn->meta_context_base_allocation = false;
+    memcpy (conn->exportname, buf, exportnamelen);
+    conn->exportname[exportnamelen] = '\0';
+    conn->exportnamelen = exportnamelen;
+  }
+  debug ("newstyle negotiation: %s: client requested export '%.*s'",
+         name_of_nbd_opt (option), (int) exportnamelen, buf);
+  return 0;
+}
+
 static int
 negotiate_handshake_newstyle_options (struct connection *conn)
 {
-  struct new_option new_option;
+  struct nbd_new_option new_option;
   size_t nr_options;
   uint64_t version;
   uint32_t option;
   uint32_t optlen;
-  char data[MAX_OPTION_LENGTH+1];
-  struct new_handshake_finish handshake_finish;
+  struct nbd_export_name_option_reply handshake_finish;
   const char *optname;
+  uint64_t exportsize;
 
   for (nr_options = 0; nr_options < MAX_NR_OPTIONS; ++nr_options) {
+    CLEANUP_FREE char *data = NULL;
+
     if (conn_recv_full (conn, &new_option, sizeof new_option,
                         "reading option: conn->recv: %m") == -1)
       return -1;
 
     version = be64toh (new_option.version);
-    if (version != NEW_VERSION) {
+    if (version != NBD_NEW_VERSION) {
       nbdkit_error ("unknown option version %" PRIx64
                     ", expecting %" PRIx64,
-                    version, NEW_VERSION);
+                    version, NBD_NEW_VERSION);
       return -1;
     }
 
@@ -251,14 +284,29 @@ negotiate_handshake_newstyle_options (struct connection *conn)
      * of the option type.
      */
     optlen = be32toh (new_option.optlen);
-    if (optlen > MAX_OPTION_LENGTH) {
+    if (optlen > MAX_REQUEST_SIZE) {
       nbdkit_error ("client option data too long (%" PRIu32 ")", optlen);
       return -1;
     }
+    data = malloc (optlen + 1); /* Allowing a trailing NUL helps some uses */
+    if (data == NULL) {
+      nbdkit_error ("malloc: %m");
+      return -1;
+    }
 
     option = be32toh (new_option.option);
     optname = name_of_nbd_opt (option);
 
+    /* If the client lacks fixed newstyle support, it should only send
+     * NBD_OPT_EXPORT_NAME.
+     */
+    if (!(conn->cflags & NBD_FLAG_FIXED_NEWSTYLE) &&
+        option != NBD_OPT_EXPORT_NAME) {
+      if (send_newstyle_option_reply (conn, option, NBD_REP_ERR_INVALID))
+        return -1;
+      continue;
+    }
+
     /* In --tls=require / FORCEDTLS mode the only options allowed
      * before TLS negotiation are NBD_OPT_ABORT and NBD_OPT_STARTTLS.
      */
@@ -274,25 +322,24 @@ negotiate_handshake_newstyle_options (struct connection *conn)
       if (conn_recv_full (conn, data, optlen,
                           "read: %s: %m", name_of_nbd_opt (option)) == -1)
         return -1;
-      /* Apart from printing it, ignore the export name. */
-      data[optlen] = '\0';
-      debug ("newstyle negotiation: %s: "
-             "client requested export '%s' (ignored)",
-             name_of_nbd_opt (option), data);
+      if (check_export_name (conn, option, data, optlen, optlen, true) == -1)
+        return -1;
 
-      /* We have to finish the handshake by sending handshake_finish. */
-      if (finish_newstyle_options (conn) == -1)
+      /* We have to finish the handshake by sending handshake_finish.
+       * On failure, we have to disconnect.
+       */
+      if (finish_newstyle_options (conn, &exportsize) == -1)
         return -1;
 
       memset (&handshake_finish, 0, sizeof handshake_finish);
-      handshake_finish.exportsize = htobe64 (conn->exportsize);
+      handshake_finish.exportsize = htobe64 (exportsize);
       handshake_finish.eflags = htobe16 (conn->eflags);
 
       if (conn->send (conn,
                       &handshake_finish,
                       (conn->cflags & NBD_FLAG_NO_ZEROES)
-                      ? offsetof (struct new_handshake_finish, zeroes)
-                      : sizeof handshake_finish) == -1) {
+                      ? offsetof (struct nbd_export_name_option_reply, zeroes)
+                      : sizeof handshake_finish, 0) == -1) {
         nbdkit_error ("write: %s: %m", optname);
         return -1;
       }
@@ -319,8 +366,8 @@ negotiate_handshake_newstyle_options (struct connection *conn)
       /* Send back the exportname. */
       debug ("newstyle negotiation: %s: advertising export '%s'",
              name_of_nbd_opt (option), exportname);
-      if (send_newstyle_option_reply_exportname (conn, option, NBD_REP_SERVER,
-                                                 exportname) == -1)
+      if (send_newstyle_option_reply_exportname (conn, option,
+                                                 NBD_REP_SERVER) == -1)
         return -1;
 
       if (send_newstyle_option_reply (conn, option, NBD_REP_ACK) == -1)
@@ -390,7 +437,6 @@ negotiate_handshake_newstyle_options (struct connection *conn)
         uint16_t nrinfos;
         uint16_t info;
         size_t i;
-        CLEANUP_FREE char *requested_exportname = NULL;
 
         /* Validate the name length and number of INFO requests. */
         memcpy (&exportnamelen, &data[0], 4);
@@ -413,31 +459,41 @@ negotiate_handshake_newstyle_options (struct connection *conn)
           continue;
         }
 
-        /* As with NBD_OPT_EXPORT_NAME we print the export name and then
-         * ignore it.
+        /* As with NBD_OPT_EXPORT_NAME we print the export name and
+         * save it in the connection.  If an earlier
+         * NBD_OPT_SET_META_CONTEXT used an export name, it must match
+         * or else we drop the support for that context.
          */
-        requested_exportname = malloc (exportnamelen+1);
-        if (requested_exportname == NULL) {
-          nbdkit_error ("malloc: %m");
-          return -1;
+        if (check_export_name (conn, option, &data[4], exportnamelen,
+                               optlen - 6, true) == -1) {
+          if (send_newstyle_option_reply (conn, option, NBD_REP_ERR_INVALID)
+              == -1)
+            return -1;
+          continue;
         }
-        memcpy (requested_exportname, &data[4], exportnamelen);
-        requested_exportname[exportnamelen] = '\0';
-        debug ("newstyle negotiation: %s: "
-               "client requested export '%s' (ignored)",
-               optname, requested_exportname);
 
         /* The spec is confusing, but it is required that we send back
          * NBD_INFO_EXPORT, even if the client did not request it!
          * qemu client in particular does not request this, but will
-         * fail if we don't send it.
+         * fail if we don't send it.  Note that if .open fails, but we
+         * succeed at .close, then we merely return an error to the
+         * client and let them try another NBD_OPT, rather than
+         * disconnecting.
          */
-        if (finish_newstyle_options (conn) == -1)
-          return -1;
+        if (finish_newstyle_options (conn, &exportsize) == -1) {
+          if (backend_finalize (backend, conn) == -1)
+            return -1;
+          backend_close (backend, conn);
+          if (send_newstyle_option_reply (conn, option,
+                                          NBD_REP_ERR_UNKNOWN) == -1)
+            return -1;
+          continue;
+        }
 
         if (send_newstyle_option_reply_info_export (conn, option,
                                                     NBD_REP_INFO,
-                                                    NBD_INFO_EXPORT) == -1)
+                                                    NBD_INFO_EXPORT,
+                                                    exportsize) == -1)
           return -1;
 
         /* For now we ignore all other info requests (but we must
@@ -460,11 +516,17 @@ negotiate_handshake_newstyle_options (struct connection *conn)
       }
 
       /* Unlike NBD_OPT_EXPORT_NAME, NBD_OPT_GO sends back an ACK
-       * or ERROR packet.
+       * or ERROR packet.  If this was NBD_OPT_LIST, call .close.
        */
       if (send_newstyle_option_reply (conn, option, NBD_REP_ACK) == -1)
         return -1;
 
+      if (option == NBD_OPT_INFO) {
+        if (backend_finalize (backend, conn) == -1)
+          return -1;
+        backend_close (backend, conn);
+      }
+
       break;
 
     case NBD_OPT_STRUCTURED_REPLY:
@@ -481,6 +543,17 @@ negotiate_handshake_newstyle_options (struct connection *conn)
       debug ("newstyle negotiation: %s: client requested structured replies",
              name_of_nbd_opt (option));
 
+      if (no_sr) {
+        /* Must fail with ERR_UNSUP for qemu 4.2 to remain happy;
+         * but failing with ERR_POLICY would have been nicer.
+         */
+        if (send_newstyle_option_reply (conn, option, NBD_REP_ERR_UNSUP) == -1)
+          return -1;
+        debug ("newstyle negotiation: %s: structured replies are disabled",
+               name_of_nbd_opt (option));
+        break;
+      }
+
       if (send_newstyle_option_reply (conn, option, NBD_REP_ACK) == -1)
         return -1;
 
@@ -499,6 +572,9 @@ negotiate_handshake_newstyle_options (struct connection *conn)
         if (conn_recv_full (conn, data, optlen, "read: %s: %m", optname) == -1)
           return -1;
 
+        /* Note that we support base:allocation whether or not the plugin
+         * supports can_extents.
+         */
         if (!conn->structured_replies) {
           if (send_newstyle_option_reply (conn, option, NBD_REP_ERR_INVALID)
               == -1)
@@ -523,9 +599,17 @@ negotiate_handshake_newstyle_options (struct connection *conn)
           continue;
         }
 
-        /* Discard the export name. */
+        /* Remember the export name: the NBD spec says that if the client
+         * later uses NBD_OPT_GO on a different export, then the context
+         * returned here is not usable.
+         */
         memcpy (&exportnamelen, &data[0], 4);
         exportnamelen = be32toh (exportnamelen);
+        what = "validating export name";
+        if (check_export_name (conn, option, &data[4], exportnamelen,
+                               optlen - 8,
+                               option == NBD_OPT_SET_META_CONTEXT) == -1)
+          goto opt_meta_invalid_option_len;
         opt_index = 4 + exportnamelen;
 
         /* Read the number of queries. */
@@ -542,10 +626,10 @@ negotiate_handshake_newstyle_options (struct connection *conn)
         debug ("newstyle negotiation: %s: %s count: %d", optname,
                option == NBD_OPT_LIST_META_CONTEXT ? "query" : "set",
                nr_queries);
+        if (option == NBD_OPT_SET_META_CONTEXT)
+          conn->meta_context_base_allocation = false;
         if (nr_queries == 0) {
-          if (option == NBD_OPT_SET_META_CONTEXT)
-            conn->meta_context_base_allocation = false;
-          else /* LIST */ {
+          if (option == NBD_OPT_LIST_META_CONTEXT) {
             if (send_newstyle_option_reply_meta_context
                 (conn, option, NBD_REP_META_CONTEXT,
                  0, "base:allocation") == -1)
@@ -565,7 +649,8 @@ negotiate_handshake_newstyle_options (struct connection *conn)
             querylen = be32toh (querylen);
             opt_index += 4;
             what = "reading query string";
-            if (opt_index + querylen > optlen)
+            if (check_string (option, &data[opt_index], querylen,
+                              optlen - opt_index, "meta context query") == -1)
               goto opt_meta_invalid_option_len;
 
             debug ("newstyle negotiation: %s: %s %.*s",
@@ -645,18 +730,18 @@ negotiate_handshake_newstyle_options (struct connection *conn)
 int
 protocol_handshake_newstyle (struct connection *conn)
 {
-  struct new_handshake handshake;
+  struct nbd_new_handshake handshake;
   uint16_t gflags;
 
-  gflags = NBD_FLAG_FIXED_NEWSTYLE | NBD_FLAG_NO_ZEROES;
+  gflags = (NBD_FLAG_FIXED_NEWSTYLE | NBD_FLAG_NO_ZEROES) & mask_handshake;
 
   debug ("newstyle negotiation: flags: global 0x%x", gflags);
 
-  memcpy (handshake.nbdmagic, "NBDMAGIC", 8);
-  handshake.version = htobe64 (NEW_VERSION);
+  handshake.nbdmagic = htobe64 (NBD_MAGIC);
+  handshake.version = htobe64 (NBD_NEW_VERSION);
   handshake.gflags = htobe16 (gflags);
 
-  if (conn->send (conn, &handshake, sizeof handshake) == -1) {
+  if (conn->send (conn, &handshake, sizeof handshake, 0) == -1) {
     nbdkit_error ("write: %s: %m", "sending newstyle handshake");
     return -1;
   }
@@ -669,7 +754,7 @@ protocol_handshake_newstyle (struct connection *conn)
   /* ... which we check for accuracy. */
   debug ("newstyle negotiation: client flags: 0x%x", conn->cflags);
   if (conn->cflags & ~gflags) {
-    nbdkit_error ("client requested unknown flags 0x%x", conn->cflags);
+    nbdkit_error ("client requested unexpected flags 0x%x", conn->cflags);
     return -1;
   }
 
