@@ -45,18 +45,8 @@
 #include "internal.h"
 #include "byte-swapping.h"
 #include "minmax.h"
-#include "protocol.h"
-
-/* Maximum read or write request that we will handle. */
-#define MAX_REQUEST_SIZE (64 * 1024 * 1024)
-
-static bool
-valid_range (struct connection *conn, uint64_t offset, uint32_t count)
-{
-  uint64_t exportsize = conn->exportsize;
-
-  return count > 0 && offset <= exportsize && offset + count <= exportsize;
-}
+#include "nbd-protocol.h"
+#include "protostrings.h"
 
 static bool
 validate_request (struct connection *conn,
@@ -64,7 +54,7 @@ validate_request (struct connection *conn,
                   uint32_t *error)
 {
   /* Readonly connection? */
-  if (conn->readonly &&
+  if (conn->eflags & NBD_FLAG_READ_ONLY &&
       (cmd == NBD_CMD_WRITE || cmd == NBD_CMD_TRIM ||
        cmd == NBD_CMD_WRITE_ZEROES)) {
     nbdkit_error ("invalid request: %s: write request on readonly connection",
@@ -76,11 +66,12 @@ validate_request (struct connection *conn,
   /* Validate cmd, offset, count. */
   switch (cmd) {
   case NBD_CMD_READ:
+  case NBD_CMD_CACHE:
   case NBD_CMD_WRITE:
   case NBD_CMD_TRIM:
   case NBD_CMD_WRITE_ZEROES:
   case NBD_CMD_BLOCK_STATUS:
-    if (!valid_range (conn, offset, count)) {
+    if (!backend_valid_range (backend, conn, offset, count)) {
       /* XXX Allow writes to extend the disk? */
       nbdkit_error ("invalid request: %s: offset and count are out of range: "
                     "offset=%" PRIu64 " count=%" PRIu32,
@@ -109,7 +100,8 @@ validate_request (struct connection *conn,
 
   /* Validate flags */
   if (flags & ~(NBD_CMD_FLAG_FUA | NBD_CMD_FLAG_NO_HOLE |
-                NBD_CMD_FLAG_DF | NBD_CMD_FLAG_REQ_ONE)) {
+                NBD_CMD_FLAG_DF | NBD_CMD_FLAG_REQ_ONE |
+                NBD_CMD_FLAG_FAST_ZERO)) {
     nbdkit_error ("invalid request: unknown flag (0x%x)", flags);
     *error = EINVAL;
     return false;
@@ -120,6 +112,13 @@ validate_request (struct connection *conn,
     *error = EINVAL;
     return false;
   }
+  if ((flags & NBD_CMD_FLAG_FAST_ZERO) &&
+      cmd != NBD_CMD_WRITE_ZEROES) {
+    nbdkit_error ("invalid request: "
+                  "FAST_ZERO flag needs WRITE_ZEROES request");
+    *error = EINVAL;
+    return false;
+  }
   if (flags & NBD_CMD_FLAG_DF) {
     if (cmd != NBD_CMD_READ) {
       nbdkit_error ("invalid request: DF flag needs READ request");
@@ -140,7 +139,7 @@ validate_request (struct connection *conn,
     *error = EINVAL;
     return false;
   }
-  if (!conn->can_fua && (flags & NBD_CMD_FLAG_FUA)) {
+  if (flags & NBD_CMD_FLAG_FUA && !(conn->eflags & NBD_FLAG_SEND_FUA)) {
     nbdkit_error ("invalid request: FUA flag not supported");
     *error = EINVAL;
     return false;
@@ -157,7 +156,7 @@ validate_request (struct connection *conn,
   }
 
   /* Flush allowed? */
-  if (!conn->can_flush && cmd == NBD_CMD_FLUSH) {
+  if (cmd == NBD_CMD_FLUSH && !(conn->eflags & NBD_FLAG_SEND_FLUSH)) {
     nbdkit_error ("invalid request: %s: flush operation not supported",
                   name_of_nbd_cmd (cmd));
     *error = EINVAL;
@@ -165,7 +164,7 @@ validate_request (struct connection *conn,
   }
 
   /* Trim allowed? */
-  if (!conn->can_trim && cmd == NBD_CMD_TRIM) {
+  if (cmd == NBD_CMD_TRIM && !(conn->eflags & NBD_FLAG_SEND_TRIM)) {
     nbdkit_error ("invalid request: %s: trim operation not supported",
                   name_of_nbd_cmd (cmd));
     *error = EINVAL;
@@ -173,13 +172,22 @@ validate_request (struct connection *conn,
   }
 
   /* Zero allowed? */
-  if (!conn->can_zero && cmd == NBD_CMD_WRITE_ZEROES) {
+  if (cmd == NBD_CMD_WRITE_ZEROES &&
+      !(conn->eflags & NBD_FLAG_SEND_WRITE_ZEROES)) {
     nbdkit_error ("invalid request: %s: write zeroes operation not supported",
                   name_of_nbd_cmd (cmd));
     *error = EINVAL;
     return false;
   }
 
+  /* Cache allowed? */
+  if (cmd == NBD_CMD_CACHE && !(conn->eflags & NBD_FLAG_SEND_CACHE)) {
+    nbdkit_error ("invalid request: %s: cache operation not supported",
+                  name_of_nbd_cmd (cmd));
+    *error = EINVAL;
+    return false;
+  }
+
   /* Block status allowed? */
   if (cmd == NBD_CMD_BLOCK_STATUS) {
     if (!conn->structured_replies) {
@@ -222,7 +230,6 @@ handle_request (struct connection *conn,
                 void *buf, struct nbdkit_extents *extents)
 {
   uint32_t f = 0;
-  bool fua = conn->can_fua && (flags & NBD_CMD_FLAG_FUA);
   int err = 0;
 
   /* Clear the error, so that we know if the plugin calls
@@ -231,63 +238,51 @@ handle_request (struct connection *conn,
 
   switch (cmd) {
   case NBD_CMD_READ:
-    if (backend->pread (backend, conn, buf, count, offset, 0, &err) == -1)
+    if (backend_pread (backend, conn, buf, count, offset, 0, &err) == -1)
       return err;
     break;
 
   case NBD_CMD_WRITE:
-    if (fua)
+    if (flags & NBD_CMD_FLAG_FUA)
       f |= NBDKIT_FLAG_FUA;
-    if (backend->pwrite (backend, conn, buf, count, offset, f, &err) == -1)
+    if (backend_pwrite (backend, conn, buf, count, offset, f, &err) == -1)
       return err;
     break;
 
   case NBD_CMD_FLUSH:
-    if (backend->flush (backend, conn, 0, &err) == -1)
+    if (backend_flush (backend, conn, 0, &err) == -1)
       return err;
     break;
 
   case NBD_CMD_TRIM:
-    if (fua)
+    if (flags & NBD_CMD_FLAG_FUA)
       f |= NBDKIT_FLAG_FUA;
-    if (backend->trim (backend, conn, count, offset, f, &err) == -1)
+    if (backend_trim (backend, conn, count, offset, f, &err) == -1)
+      return err;
+    break;
+
+  case NBD_CMD_CACHE:
+    if (backend_cache (backend, conn, count, offset, 0, &err) == -1)
       return err;
     break;
 
   case NBD_CMD_WRITE_ZEROES:
     if (!(flags & NBD_CMD_FLAG_NO_HOLE))
       f |= NBDKIT_FLAG_MAY_TRIM;
-    if (fua)
+    if (flags & NBD_CMD_FLAG_FUA)
       f |= NBDKIT_FLAG_FUA;
-    if (backend->zero (backend, conn, count, offset, f, &err) == -1)
+    if (flags & NBD_CMD_FLAG_FAST_ZERO)
+      f |= NBDKIT_FLAG_FAST_ZERO;
+    if (backend_zero (backend, conn, count, offset, f, &err) == -1)
       return err;
     break;
 
   case NBD_CMD_BLOCK_STATUS:
-    /* The other backend methods don't check can_*.  That is because
-     * those methods are implicitly suppressed by returning eflags to
-     * the client.  However there is no eflag for extents so we must
-     * check it here.
-     */
-    if (conn->can_extents) {
-      if (flags & NBD_CMD_FLAG_REQ_ONE)
-        f |= NBDKIT_FLAG_REQ_ONE;
-      if (backend->extents (backend, conn, count, offset, f,
-                            extents, &err) == -1)
-        return err;
-    }
-    else {
-      int r;
-
-      /* By default it is safe assume that everything in the range is
-       * allocated.
-       */
-      errno = 0;
-      r = nbdkit_add_extent (extents, offset, count, 0 /* allocated data */);
-      if (r == -1)
-        return errno ? errno : EINVAL;
-      return 0;
-    }
+    if (flags & NBD_CMD_FLAG_REQ_ONE)
+      f |= NBDKIT_FLAG_REQ_ONE;
+    if (backend_extents (backend, conn, count, offset, f,
+                         extents, &err) == -1)
+      return err;
     break;
 
   default:
@@ -326,7 +321,7 @@ skip_over_write_buffer (int sock, size_t count)
 
 /* Convert a system errno to an NBD_E* error code. */
 static int
-nbd_errno (int error, bool flag_df)
+nbd_errno (int error, uint16_t flags)
 {
   switch (error) {
   case 0:
@@ -348,10 +343,17 @@ nbd_errno (int error, bool flag_df)
   case ESHUTDOWN:
     return NBD_ESHUTDOWN;
 #endif
+  case ENOTSUP:
+#if ENOTSUP != EOPNOTSUPP
+  case EOPNOTSUPP:
+#endif
+    if (flags & NBD_CMD_FLAG_FAST_ZERO)
+      return NBD_ENOTSUP;
+    return NBD_EINVAL;
   case EOVERFLOW:
-    if (flag_df)
+    if (flags & NBD_CMD_FLAG_DF)
       return NBD_EOVERFLOW;
-    /* fallthrough */
+    return NBD_EINVAL;
   case EINVAL:
   default:
     return NBD_EINVAL;
@@ -360,19 +362,20 @@ nbd_errno (int error, bool flag_df)
 
 static int
 send_simple_reply (struct connection *conn,
-                   uint64_t handle, uint16_t cmd,
+                   uint64_t handle, uint16_t cmd, uint16_t flags,
                    const char *buf, uint32_t count,
                    uint32_t error)
 {
   ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&conn->write_lock);
-  struct simple_reply reply;
+  struct nbd_simple_reply reply;
   int r;
+  int f = (cmd == NBD_CMD_READ && !error) ? SEND_MORE : 0;
 
   reply.magic = htobe32 (NBD_SIMPLE_REPLY_MAGIC);
   reply.handle = handle;
-  reply.error = htobe32 (nbd_errno (error, false));
+  reply.error = htobe32 (nbd_errno (error, flags));
 
-  r = conn->send (conn, &reply, sizeof reply);
+  r = conn->send (conn, &reply, sizeof reply, f);
   if (r == -1) {
     nbdkit_error ("write reply: %s: %m", name_of_nbd_cmd (cmd));
     return connection_set_status (conn, -1);
@@ -380,7 +383,7 @@ send_simple_reply (struct connection *conn,
 
   /* Send the read data buffer. */
   if (cmd == NBD_CMD_READ && !error) {
-    r = conn->send (conn, buf, count);
+    r = conn->send (conn, buf, count, 0);
     if (r == -1) {
       nbdkit_error ("write data: %s: %m", name_of_nbd_cmd (cmd));
       return connection_set_status (conn, -1);
@@ -401,8 +404,8 @@ send_structured_reply_read (struct connection *conn,
    * that yet we acquire the lock for the whole function.
    */
   ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&conn->write_lock);
-  struct structured_reply reply;
-  struct structured_reply_offset_data offset_data;
+  struct nbd_structured_reply reply;
+  struct nbd_structured_reply_offset_data offset_data;
   int r;
 
   assert (cmd == NBD_CMD_READ);
@@ -413,7 +416,7 @@ send_structured_reply_read (struct connection *conn,
   reply.type = htobe16 (NBD_REPLY_TYPE_OFFSET_DATA);
   reply.length = htobe32 (count + sizeof offset_data);
 
-  r = conn->send (conn, &reply, sizeof reply);
+  r = conn->send (conn, &reply, sizeof reply, SEND_MORE);
   if (r == -1) {
     nbdkit_error ("write reply: %s: %m", name_of_nbd_cmd (cmd));
     return connection_set_status (conn, -1);
@@ -421,13 +424,13 @@ send_structured_reply_read (struct connection *conn,
 
   /* Send the offset + read data buffer. */
   offset_data.offset = htobe64 (offset);
-  r = conn->send (conn, &offset_data, sizeof offset_data);
+  r = conn->send (conn, &offset_data, sizeof offset_data, SEND_MORE);
   if (r == -1) {
     nbdkit_error ("write data: %s: %m", name_of_nbd_cmd (cmd));
     return connection_set_status (conn, -1);
   }
 
-  r = conn->send (conn, buf, count);
+  r = conn->send (conn, buf, count, 0);
   if (r == -1) {
     nbdkit_error ("write data: %s: %m", name_of_nbd_cmd (cmd));
     return connection_set_status (conn, -1);
@@ -439,7 +442,7 @@ send_structured_reply_read (struct connection *conn,
 /* Convert a list of extents into NBD_REPLY_TYPE_BLOCK_STATUS blocks.
  * The rules here are very complicated.  Read the spec carefully!
  */
-static struct block_descriptor *
+static struct nbd_block_descriptor *
 extents_to_block_descriptors (struct nbdkit_extents *extents,
                               uint16_t flags,
                               uint32_t count, uint64_t offset,
@@ -448,13 +451,14 @@ extents_to_block_descriptors (struct nbdkit_extents *extents,
   const bool req_one = flags & NBD_CMD_FLAG_REQ_ONE;
   const size_t nr_extents = nbdkit_extents_count (extents);
   size_t i;
-  struct block_descriptor *blocks;
+  struct nbd_block_descriptor *blocks;
 
   /* This is checked in server/plugins.c. */
   assert (nr_extents >= 1);
 
   /* We may send fewer than nr_extents blocks, but never more. */
-  blocks = calloc (req_one ? 1 : nr_extents, sizeof (struct block_descriptor));
+  blocks = calloc (req_one ? 1 : nr_extents,
+                   sizeof (struct nbd_block_descriptor));
   if (blocks == NULL) {
     nbdkit_error ("calloc: %m");
     return NULL;
@@ -504,8 +508,8 @@ extents_to_block_descriptors (struct nbdkit_extents *extents,
 
 #if 0
   for (i = 0; i < *nr_blocks; ++i)
-    nbdkit_debug ("block status: sending block %" PRIu32 " type %" PRIu32,
-                  blocks[i].length, blocks[i].status_flags);
+    debug ("block status: sending block %" PRIu32 " type %" PRIu32,
+           blocks[i].length, blocks[i].status_flags);
 #endif
 
   /* Convert to big endian for the protocol. */
@@ -525,8 +529,8 @@ send_structured_reply_block_status (struct connection *conn,
                                     struct nbdkit_extents *extents)
 {
   ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&conn->write_lock);
-  struct structured_reply reply;
-  CLEANUP_FREE struct block_descriptor *blocks = NULL;
+  struct nbd_structured_reply reply;
+  CLEANUP_FREE struct nbd_block_descriptor *blocks = NULL;
   size_t nr_blocks;
   uint32_t context_id;
   size_t i;
@@ -545,9 +549,9 @@ send_structured_reply_block_status (struct connection *conn,
   reply.flags = htobe16 (NBD_REPLY_FLAG_DONE);
   reply.type = htobe16 (NBD_REPLY_TYPE_BLOCK_STATUS);
   reply.length = htobe32 (sizeof context_id +
-                          nr_blocks * sizeof (struct block_descriptor));
+                          nr_blocks * sizeof (struct nbd_block_descriptor));
 
-  r = conn->send (conn, &reply, sizeof reply);
+  r = conn->send (conn, &reply, sizeof reply, SEND_MORE);
   if (r == -1) {
     nbdkit_error ("write reply: %s: %m", name_of_nbd_cmd (cmd));
     return connection_set_status (conn, -1);
@@ -555,7 +559,7 @@ send_structured_reply_block_status (struct connection *conn,
 
   /* Send the base:allocation context ID. */
   context_id = htobe32 (base_allocation_id);
-  r = conn->send (conn, &context_id, sizeof context_id);
+  r = conn->send (conn, &context_id, sizeof context_id, SEND_MORE);
   if (r == -1) {
     nbdkit_error ("write reply: %s: %m", name_of_nbd_cmd (cmd));
     return connection_set_status (conn, -1);
@@ -563,7 +567,8 @@ send_structured_reply_block_status (struct connection *conn,
 
   /* Send each block descriptor. */
   for (i = 0; i < nr_blocks; ++i) {
-    r = conn->send (conn, &blocks[i], sizeof blocks[i]);
+    r = conn->send (conn, &blocks[i], sizeof blocks[i],
+                    i == nr_blocks - 1 ? 0 : SEND_MORE);
     if (r == -1) {
       nbdkit_error ("write reply: %s: %m", name_of_nbd_cmd (cmd));
       return connection_set_status (conn, -1);
@@ -579,8 +584,8 @@ send_structured_reply_error (struct connection *conn,
                              uint32_t error)
 {
   ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&conn->write_lock);
-  struct structured_reply reply;
-  struct structured_reply_error error_data;
+  struct nbd_structured_reply reply;
+  struct nbd_structured_reply_error error_data;
   int r;
 
   reply.magic = htobe32 (NBD_STRUCTURED_REPLY_MAGIC);
@@ -589,16 +594,16 @@ send_structured_reply_error (struct connection *conn,
   reply.type = htobe16 (NBD_REPLY_TYPE_ERROR);
   reply.length = htobe32 (0 /* no human readable error */ + sizeof error_data);
 
-  r = conn->send (conn, &reply, sizeof reply);
+  r = conn->send (conn, &reply, sizeof reply, SEND_MORE);
   if (r == -1) {
     nbdkit_error ("write error reply: %m");
     return connection_set_status (conn, -1);
   }
 
   /* Send the error. */
-  error_data.error = htobe32 (nbd_errno (error, flags & NBD_CMD_FLAG_DF));
+  error_data.error = htobe32 (nbd_errno (error, flags));
   error_data.len = htobe16 (0);
-  r = conn->send (conn, &error_data, sizeof error_data);
+  r = conn->send (conn, &error_data, sizeof error_data, 0);
   if (r == -1) {
     nbdkit_error ("write data: %s: %m", name_of_nbd_cmd (cmd));
     return connection_set_status (conn, -1);
@@ -612,7 +617,7 @@ int
 protocol_recv_request_send_reply (struct connection *conn)
 {
   int r;
-  struct request request;
+  struct nbd_request request;
   uint16_t cmd, flags;
   uint32_t magic, count, error = 0;
   uint64_t offset;
@@ -677,7 +682,7 @@ protocol_recv_request_send_reply (struct connection *conn)
 
     /* Allocate the extents list for block status only. */
     if (cmd == NBD_CMD_BLOCK_STATUS) {
-      extents = nbdkit_extents_new (offset, conn->exportsize);
+      extents = nbdkit_extents_new (offset, backend_get_size (backend, conn));
       if (extents == NULL) {
         error = ENOMEM;
         goto send_reply;
@@ -746,5 +751,6 @@ protocol_recv_request_send_reply (struct connection *conn)
                                           error);
   }
   else
-    return send_simple_reply (conn, request.handle, cmd, buf, count, error);
+    return send_simple_reply (conn, request.handle, cmd, flags, buf, count,
+                              error);
 }
