@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2013-2018 Red Hat Inc.
+ * Copyright (C) 2013-2019 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -35,15 +35,22 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdbool.h>
+#include <inttypes.h>
 #include <string.h>
 #include <unistd.h>
-#include <sys/types.h>
-#include <sys/socket.h>
-#include <sys/un.h>
-#include <netdb.h>
 #include <poll.h>
 #include <errno.h>
 #include <assert.h>
+#include <sys/types.h>
+#include <sys/socket.h>
+#include <sys/un.h>
+#include <netinet/in.h>
+#include <netinet/tcp.h>
+#include <netdb.h>
+
+#ifdef HAVE_LINUX_VM_SOCKETS_H
+#include <linux/vm_sockets.h>
+#endif
 
 #ifdef HAVE_LIBSELINUX
 #include <selinux/selinux.h>
@@ -52,6 +59,7 @@
 #include <pthread.h>
 
 #include "internal.h"
+#include "utils.h"
 
 static void
 set_selinux_label (void)
@@ -105,7 +113,14 @@ bind_unix_socket (size_t *nr_socks)
 
   set_selinux_label ();
 
+#ifdef SOCK_CLOEXEC
   sock = socket (AF_UNIX, SOCK_STREAM|SOCK_CLOEXEC, 0);
+#else
+  /* Fortunately, this code is only run at startup, so there is no
+   * risk of the fd leaking to a plugin's fork()
+   */
+  sock = set_cloexec (socket (AF_UNIX, SOCK_STREAM, 0));
+#endif
   if (sock == -1) {
     perror ("bind_unix_socket: socket");
     exit (EXIT_FAILURE);
@@ -173,7 +188,14 @@ bind_tcpip_socket (size_t *nr_socks)
 
     set_selinux_label ();
 
-    sock = socket (a->ai_family, a->ai_socktype, a->ai_protocol);
+#ifdef SOCK_CLOEXEC
+    sock = socket (a->ai_family, a->ai_socktype | SOCK_CLOEXEC, a->ai_protocol);
+#else
+    /* Fortunately, this code is only run at startup, so there is no
+     * risk of the fd leaking to a plugin's fork()
+     */
+    sock = set_cloexec (socket (a->ai_family, a->ai_socktype, a->ai_protocol));
+#endif
     if (sock == -1) {
       perror ("bind_tcpip_socket: socket");
       exit (EXIT_FAILURE);
@@ -230,21 +252,90 @@ bind_tcpip_socket (size_t *nr_socks)
   return socks;
 }
 
-void
-free_listening_sockets (int *socks, size_t nr_socks)
+int *
+bind_vsock (size_t *nr_socks)
 {
-  size_t i;
+#ifdef AF_VSOCK
+  uint32_t vsock_port;
+  int sock;
+  int *ret;
+  struct sockaddr_vm addr;
 
-  for (i = 0; i < nr_socks; ++i)
-    close (socks[i]);
-  free (socks);
+  if (port == NULL)
+    vsock_port = 10809;
+  else {
+    /* --port parameter must be numeric for vsock, unless
+     * /etc/services is extended but that seems unlikely. XXX
+     */
+    if (nbdkit_parse_uint32_t ("port", port, &vsock_port) == -1)
+      exit (EXIT_FAILURE);
+  }
+
+  /* Any platform with AF_VSOCK also supports SOCK_CLOEXEC so there is
+   * no fallback path.
+   */
+  sock = socket (AF_VSOCK, SOCK_STREAM|SOCK_CLOEXEC, 0);
+  if (sock == -1) {
+    perror ("bind_vsock: socket");
+    exit (EXIT_FAILURE);
+  }
+
+  memset (&addr, 0, sizeof addr);
+  addr.svm_family = AF_VSOCK;
+  addr.svm_cid = VMADDR_CID_ANY;
+  addr.svm_port = vsock_port;
+
+  if (bind (sock, (struct sockaddr *) &addr, sizeof addr) == -1) {
+    perror (unixsocket);
+    exit (EXIT_FAILURE);
+  }
+
+  if (listen (sock, SOMAXCONN) == -1) {
+    perror ("listen");
+    exit (EXIT_FAILURE);
+  }
+
+  ret = malloc (sizeof (int));
+  if (!ret) {
+    perror ("malloc");
+    exit (EXIT_FAILURE);
+  }
+  ret[0] = sock;
+  *nr_socks = 1;
+
+  /* It's not easy to get the actual CID here.
+   * IOCTL_VM_SOCKETS_GET_LOCAL_CID is documented, but requires
+   * opening /dev/vsock which is not accessible to non-root users.
+   * bind above doesn't update the sockaddr.  Using getsockname
+   * doesn't work.
+   */
+  debug ("bound to vsock any:%" PRIu32, addr.svm_port);
+
+  return ret;
+
+#else
+  /* Can't happen because main() checks if AF_VSOCK is defined and
+   * prevents vsock from being set, so this function can never be
+   * called.
+   */
+  abort ();
+#endif
 }
 
+/* This counts the number of connection threads running (note: not the
+ * number of worker threads, each connection thread will start many
+ * worker independent threads in the current implementation).  The
+ * purpose of this is so we can wait for all the connection threads to
+ * exit before we return from accept_incoming_connections, so that
+ * unload-time actions happen with no connections open.
+ */
+static pthread_mutex_t count_mutex = PTHREAD_MUTEX_INITIALIZER;
+static pthread_cond_t count_cond = PTHREAD_COND_INITIALIZER;
+static unsigned count = 0;
+
 struct thread_data {
   int sock;
   size_t instance_num;
-  struct sockaddr addr;
-  socklen_t addrlen;
 };
 
 static void *
@@ -254,14 +345,23 @@ start_thread (void *datav)
 
   debug ("accepted connection");
 
+  pthread_mutex_lock (&count_mutex);
+  count++;
+  pthread_mutex_unlock (&count_mutex);
+
   /* Set thread-local data. */
   threadlocal_new_server_thread ();
   threadlocal_set_instance_num (data->instance_num);
-  threadlocal_set_sockaddr (&data->addr, data->addrlen);
 
   handle_single_connection (data->sock, data->sock);
 
   free (data);
+
+  pthread_mutex_lock (&count_mutex);
+  count--;
+  pthread_cond_signal (&count_cond);
+  pthread_mutex_unlock (&count_mutex);
+
   return NULL;
 }
 
@@ -273,18 +373,32 @@ accept_connection (int listen_sock)
   pthread_t thread;
   struct thread_data *thread_data;
   static size_t instance_num = 1;
+  const int flag = 1;
 
   thread_data = malloc (sizeof *thread_data);
-  if (!thread_data) {
+  if (unlikely (!thread_data)) {
     perror ("malloc");
     return;
   }
 
   thread_data->instance_num = instance_num++;
-  thread_data->addrlen = sizeof thread_data->addr;
  again:
-  thread_data->sock = accept (listen_sock,
-                              &thread_data->addr, &thread_data->addrlen);
+#ifdef HAVE_ACCEPT4
+  thread_data->sock = accept4 (listen_sock, NULL, NULL, SOCK_CLOEXEC);
+#else
+  /* If we were fully parallel, then this function could be accepting
+   * connections in one thread while another thread could be in a
+   * plugin trying to fork.  But plugins.c forced thread_model to
+   * serialize_all_requests when it detects a lack of atomic CLOEXEC,
+   * at which point, we can use a mutex to ensure we aren't accepting
+   * until the plugin is not running, making non-atomicity okay.
+   */
+  assert (backend->thread_model (backend) <=
+          NBDKIT_THREAD_MODEL_SERIALIZE_ALL_REQUESTS);
+  lock_request (NULL);
+  thread_data->sock = set_cloexec (accept (listen_sock, NULL, NULL));
+  unlock_request (NULL);
+#endif
   if (thread_data->sock == -1) {
     if (errno == EINTR || errno == EAGAIN)
       goto again;
@@ -293,6 +407,11 @@ accept_connection (int listen_sock)
     return;
   }
 
+  /* Disable Nagle's algorithm on this socket.  However we don't want
+   * to fail if this doesn't work.
+   */
+  setsockopt (thread_data->sock, IPPROTO_TCP, TCP_NODELAY, &flag, sizeof flag);
+
   /* Start a thread to handle this connection.  Note we always do this
    * even for non-threaded plugins.  There are mutexes in plugins.c
    * which ensure that non-threaded plugins are handled correctly.
@@ -301,7 +420,7 @@ accept_connection (int listen_sock)
   pthread_attr_setdetachstate (&attrs, PTHREAD_CREATE_DETACHED);
   err = pthread_create (&thread, &attrs, start_thread, thread_data);
   pthread_attr_destroy (&attrs);
-  if (err != 0) {
+  if (unlikely (err != 0)) {
     fprintf (stderr, "%s: pthread_create: %s\n", program_name, strerror (err));
     close (thread_data->sock);
     free (thread_data);
@@ -325,10 +444,16 @@ accept_connection (int listen_sock)
 static void
 check_sockets_and_quit_fd (int *socks, size_t nr_socks)
 {
-  struct pollfd fds[nr_socks + 1];
   size_t i;
   int r;
 
+  CLEANUP_FREE struct pollfd *fds =
+    malloc (sizeof (struct pollfd) * (nr_socks+1));
+  if (fds == NULL) {
+    perror ("malloc");
+    exit (EXIT_FAILURE);
+  }
+
   for (i = 0; i < nr_socks; ++i) {
     fds[i].fd = socks[i];
     fds[i].events = POLLIN;
@@ -362,6 +487,26 @@ check_sockets_and_quit_fd (int *socks, size_t nr_socks)
 void
 accept_incoming_connections (int *socks, size_t nr_socks)
 {
+  size_t i;
+  int err;
+
   while (!quit)
     check_sockets_and_quit_fd (socks, nr_socks);
+
+  /* Wait for all threads to exit. */
+  pthread_mutex_lock (&count_mutex);
+  for (;;) {
+    if (count == 0)
+      break;
+    err = pthread_cond_wait (&count_cond, &count_mutex);
+    if (err != 0) {
+      errno = err;
+      perror ("pthread_cond_wait");
+    }
+  }
+  pthread_mutex_unlock (&count_mutex);
+
+  for (i = 0; i < nr_socks; ++i)
+    close (socks[i]);
+  free (socks);
 }
