@@ -2,8 +2,8 @@
 "       [-e|--exportname EXPORTNAME] [--exit-with-parent]\n"
 "       [--filter FILTER ...] [-f|--foreground]\n"
 "       [-g|--group GROUP] [-i|--ipaddr IPADDR]\n"
-"       [--log stderr|syslog]\n"
-"       [-n|--newstyle] [-o|--oldstyle]\n"
+"       [--log stderr|syslog|null]\n"
+"       [-n|--newstyle] [--mask-handshake MASK] [--no-sr] [-o|--oldstyle]\n"
 "       [-P|--pidfile PIDFILE]\n"
 "       [-p|--port PORT] [-r|--readonly]\n"
 "       [--run CMD] [-s|--single] [--selinux-label LABEL]\n"
@@ -12,7 +12,7 @@
 "       [--tls-certificates /path/to/certificates]\n"
 "       [--tls-psk /path/to/pskfile] [--tls-verify-peer]\n"
 "       [-U|--unix SOCKET] [-u|--user USER]\n"
-"       [-v|--verbose] [-V|--version]\n"
+"       [-v|--verbose] [-V|--version] [--vsock]\n"
 "       PLUGIN [[KEY=]VALUE [KEY=VALUE [...]]]\n"
 "\n"
 "nbdkit --dump-config\n"
