@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2013-2018 Red Hat Inc.
+ * Copyright (C) 2013-2019 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -55,11 +55,10 @@
 struct threadlocal {
   char *name;                   /* Can be NULL. */
   size_t instance_num;          /* Can be 0. */
-  struct sockaddr *addr;
-  socklen_t addrlen;
   int err;
   void *buffer;
   size_t buffer_size;
+  struct connection *conn;
 };
 
 static pthread_key_t threadlocal_key;
@@ -70,7 +69,6 @@ free_threadlocal (void *threadlocalv)
   struct threadlocal *threadlocal = threadlocalv;
 
   free (threadlocal->name);
-  free (threadlocal->addr);
   free (threadlocal->buffer);
   free (threadlocal);
 }
@@ -132,22 +130,6 @@ threadlocal_set_instance_num (size_t instance_num)
     threadlocal->instance_num = instance_num;
 }
 
-void
-threadlocal_set_sockaddr (const struct sockaddr *addr, socklen_t addrlen)
-{
-  struct threadlocal *threadlocal = pthread_getspecific (threadlocal_key);
-
-  if (threadlocal) {
-    free (threadlocal->addr);
-    threadlocal->addr = calloc (1, addrlen);
-    if (threadlocal->addr == NULL) {
-      perror ("calloc");
-      exit (EXIT_FAILURE);
-    }
-    memcpy (threadlocal->addr, addr, addrlen);
-  }
-}
-
 const char *
 threadlocal_get_name (void)
 {
@@ -226,3 +208,22 @@ threadlocal_buffer (size_t size)
 
   return threadlocal->buffer;
 }
+
+/* Set (or clear) the connection that is using the current thread */
+void
+threadlocal_set_conn (struct connection *conn)
+{
+  struct threadlocal *threadlocal = pthread_getspecific (threadlocal_key);
+
+  if (threadlocal)
+    threadlocal->conn = conn;
+}
+
+/* Get the connection associated with this thread, if available */
+struct connection *
+threadlocal_get_conn (void)
+{
+  struct threadlocal *threadlocal = pthread_getspecific (threadlocal_key);
+
+  return threadlocal ? threadlocal->conn : NULL;
+}
