@@ -75,7 +75,7 @@ qemu-io -f raw "nbd+unix://?socket=$sock" \
 # Get the /proc link to the cache file, and the size of it in bytes.
 fddir="/proc/$( cat $d/cache-max-size.pid )/fd"
 ls -l $fddir ||:
-fd="$fddir/$( ls -l $fddir | grep $TMPDIR | head -1 | awk '{print $9}' )"
+fd="$fddir/$( ls -l $fddir | grep $TMPDIR/ | head -1 | awk '{print $9}' )"
 stat -L $fd
 size=$(( $(stat -L -c '%b * %B' $fd) ))
 
