@@ -1,7 +1,6 @@
 #!/usr/bin/env bash
 # nbdkit
 # Copyright (C) 2014 Red Hat Inc.
-# All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions are
@@ -37,9 +36,12 @@ source ./functions.sh
 
 # Test nbdkit -f option.
 
-rm -f foreground.pid foreground.sock
+sock=`mktemp -u`
+files="foreground.pid $sock"
+rm -f $files
+cleanup_fn rm -f $files
 
-nbdkit -f -P foreground.pid -U foreground.sock example1 &
+nbdkit -f -P foreground.pid -U $sock example1 &
 bg_pid=$!
 
 # We may have to wait a short time for the pid file to appear.
@@ -55,11 +57,12 @@ if ! test -f foreground.pid; then
 fi
 
 pid="$(cat foreground.pid)"
+cleanup_fn kill $pid
 
 test "$bg_pid" -eq "$pid"
 
 # Check the socket was created (and is a socket).
-test -S foreground.sock
+test -S $sock
 
 # Kill the process.
 kill $pid
@@ -75,5 +78,3 @@ if kill -s 0 $pid; then
     echo "$0: process did not exit after sending a signal"
     exit 1
 fi
-
-rm foreground.pid foreground.sock
