@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2018 Red Hat Inc.
+ * Copyright (C) 2018-2019 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -59,8 +59,9 @@
 #include <pthread.h>
 
 #include "byte-swapping.h"
+#include "cleanup.h"
 #include "exit-with-parent.h"
-#include "protocol.h"
+#include "nbd-protocol.h"
 
 /* Declare program_name. */
 #if HAVE_DECL_PROGRAM_INVOCATION_SHORT_NAME == 1
@@ -76,6 +77,12 @@ static void log_verify_seen_in_order (const char *msg, ...)
   __attribute__((sentinel));
 static void log_free (void);
 
+static inline void
+short_sleep (void)
+{
+  sleep (2);
+}
+
 int
 main (int argc, char *argv[])
 {
@@ -85,13 +92,13 @@ main (int argc, char *argv[])
   int err;
   pthread_t thread;
   int sock;
-  struct new_handshake handshake;
+  struct nbd_new_handshake handshake;
   uint32_t cflags;
-  struct new_option option;
-  struct new_handshake_finish handshake_finish;
+  struct nbd_new_option option;
+  struct nbd_export_name_option_reply handshake_finish;
   uint16_t eflags;
-  struct request request;
-  struct simple_reply reply;
+  struct nbd_request request;
+  struct nbd_simple_reply reply;
   char data[512];
 
 #ifndef HAVE_EXIT_WITH_PARENT
@@ -100,7 +107,9 @@ main (int argc, char *argv[])
   exit (77);
 #endif
 
-  /* Socket for communicating with nbdkit. */
+  /* Socket for communicating with nbdkit. The test doesn't care about
+   * fd leaks, so we don't bother with CLOEXEC.
+   */
   if (socketpair (AF_LOCAL, SOCK_STREAM, 0, sfd) == -1) {
     perror ("socketpair");
     exit (EXIT_FAILURE);
@@ -171,8 +180,8 @@ main (int argc, char *argv[])
     perror ("recv: handshake");
     exit (EXIT_FAILURE);
   }
-  if (memcmp (handshake.nbdmagic, "NBDMAGIC", 8) != 0 ||
-      be64toh (handshake.version) != NEW_VERSION) {
+  if (be64toh (handshake.nbdmagic) != NBD_MAGIC ||
+      be64toh (handshake.version) != NBD_NEW_VERSION) {
     fprintf (stderr, "%s: unexpected NBDMAGIC or version\n",
              program_name);
     exit (EXIT_FAILURE);
@@ -186,7 +195,7 @@ main (int argc, char *argv[])
   }
 
   /* Send NBD_OPT_EXPORT_NAME with no export name. */
-  option.version = htobe64 (NEW_VERSION);
+  option.version = htobe64 (NBD_NEW_VERSION);
   option.option = htobe32 (NBD_OPT_EXPORT_NAME);
   option.optlen = htobe32 (0);
   if (send (sock, &option, sizeof option, 0) != sizeof option) {
@@ -243,7 +252,7 @@ main (int argc, char *argv[])
   }
 
   /* Sleep briefly to allow the log to catch up. */
-  sleep (1);
+  short_sleep ();
 
   /* Verify expected log messages were seen during the handshake and
    * option negotiation phases.
@@ -279,19 +288,20 @@ main (int argc, char *argv[])
      "test_layers_plugin_config_complete",
      NULL);
 
-  /* open methods called in order. */
+  /* open methods called in outer-to-inner order, but thanks to next
+   * pointer, complete in inner-to-outer order. */
   log_verify_seen_in_order
     ("testlayersfilter3: open readonly=0",
-     "filter3: test_layers_filter_open",
      "testlayersfilter2: open readonly=0",
-     "filter2: test_layers_filter_open",
      "testlayersfilter1: open readonly=0",
-     "filter1: test_layers_filter_open",
      "testlayersplugin: open readonly=0",
      "test_layers_plugin_open",
+     "filter1: test_layers_filter_open",
+     "filter2: test_layers_filter_open",
+     "filter3: test_layers_filter_open",
      NULL);
 
-  /* prepare methods called in order.
+  /* prepare methods called in inner-to-outer order.
    *
    * Note that prepare methods only exist for filters, and they must
    * be called from inner to outer (but finalize methods below are
@@ -360,6 +370,12 @@ main (int argc, char *argv[])
      "filter1: test_layers_filter_can_extents",
      "test_layers_plugin_can_extents",
      NULL);
+  log_verify_seen_in_order
+    ("filter3: test_layers_filter_can_cache",
+     "filter2: test_layers_filter_can_cache",
+     "filter1: test_layers_filter_can_cache",
+     "test_layers_plugin_can_cache",
+     NULL);
 
   fprintf (stderr, "%s: protocol connected\n", program_name);
 
@@ -389,15 +405,15 @@ main (int argc, char *argv[])
     exit (EXIT_FAILURE);
   }
 
-  sleep (1);
+  short_sleep ();
   log_verify_seen_in_order
-    ("testlayersfilter3: pread count=512 offset=0 flags=0x0",
+    ("testlayersfilter3: pread count=512 offset=0",
      "filter3: test_layers_filter_pread",
-     "testlayersfilter2: pread count=512 offset=0 flags=0x0",
+     "testlayersfilter2: pread count=512 offset=0",
      "filter2: test_layers_filter_pread",
-     "testlayersfilter1: pread count=512 offset=0 flags=0x0",
+     "testlayersfilter1: pread count=512 offset=0",
      "filter1: test_layers_filter_pread",
-     "testlayersplugin: debug: pread count=512 offset=0",
+     "testlayersplugin: pread count=512 offset=0",
      "test_layers_plugin_pread",
      NULL);
 
@@ -423,15 +439,15 @@ main (int argc, char *argv[])
     exit (EXIT_FAILURE);
   }
 
-  sleep (1);
+  short_sleep ();
   log_verify_seen_in_order
-    ("testlayersfilter3: pwrite count=512 offset=0 flags=0x0",
+    ("testlayersfilter3: pwrite count=512 offset=0 fua=0",
      "filter3: test_layers_filter_pwrite",
-     "testlayersfilter2: pwrite count=512 offset=0 flags=0x0",
+     "testlayersfilter2: pwrite count=512 offset=0 fua=0",
      "filter2: test_layers_filter_pwrite",
-     "testlayersfilter1: pwrite count=512 offset=0 flags=0x0",
+     "testlayersfilter1: pwrite count=512 offset=0 fua=0",
      "filter1: test_layers_filter_pwrite",
-     "testlayersplugin: debug: pwrite count=512 offset=0",
+     "testlayersplugin: pwrite count=512 offset=0 fua=0",
      "test_layers_plugin_pwrite",
      NULL);
 
@@ -453,15 +469,15 @@ main (int argc, char *argv[])
     exit (EXIT_FAILURE);
   }
 
-  sleep (1);
+  short_sleep ();
   log_verify_seen_in_order
-    ("testlayersfilter3: flush flags=0x0",
+    ("testlayersfilter3: flush",
      "filter3: test_layers_filter_flush",
-     "testlayersfilter2: flush flags=0x0",
+     "testlayersfilter2: flush",
      "filter2: test_layers_filter_flush",
-     "testlayersfilter1: flush flags=0x0",
+     "testlayersfilter1: flush",
      "filter1: test_layers_filter_flush",
-     "testlayersplugin: debug: flush",
+     "testlayersplugin: flush",
      "test_layers_plugin_flush",
      NULL);
 
@@ -483,15 +499,15 @@ main (int argc, char *argv[])
     exit (EXIT_FAILURE);
   }
 
-  sleep (1);
+  short_sleep ();
   log_verify_seen_in_order
-    ("testlayersfilter3: trim count=512 offset=0 flags=0x0",
+    ("testlayersfilter3: trim count=512 offset=0 fua=0",
      "filter3: test_layers_filter_trim",
-     "testlayersfilter2: trim count=512 offset=0 flags=0x0",
+     "testlayersfilter2: trim count=512 offset=0 fua=0",
      "filter2: test_layers_filter_trim",
-     "testlayersfilter1: trim count=512 offset=0 flags=0x0",
+     "testlayersfilter1: trim count=512 offset=0 fua=0",
      "filter1: test_layers_filter_trim",
-     "testlayersplugin: debug: trim count=512 offset=0",
+     "testlayersplugin: trim count=512 offset=0 fua=0",
      "test_layers_plugin_trim",
      NULL);
 
@@ -513,18 +529,48 @@ main (int argc, char *argv[])
     exit (EXIT_FAILURE);
   }
 
-  sleep (1);
+  short_sleep ();
   log_verify_seen_in_order
-    ("testlayersfilter3: zero count=512 offset=0 flags=0x1",
+    ("testlayersfilter3: zero count=512 offset=0 may_trim=1 fua=0 fast=0",
      "filter3: test_layers_filter_zero",
-     "testlayersfilter2: zero count=512 offset=0 flags=0x1",
+     "testlayersfilter2: zero count=512 offset=0 may_trim=1 fua=0 fast=0",
      "filter2: test_layers_filter_zero",
-     "testlayersfilter1: zero count=512 offset=0 flags=0x1",
+     "testlayersfilter1: zero count=512 offset=0 may_trim=1 fua=0 fast=0",
      "filter1: test_layers_filter_zero",
-     "testlayersplugin: debug: zero count=512 offset=0 may_trim=1 fua=0",
+     "testlayersplugin: zero count=512 offset=0 may_trim=1 fua=0 fast=0",
      "test_layers_plugin_zero",
      NULL);
 
+  request.type = htobe16 (NBD_CMD_CACHE);
+  request.offset = htobe64 (0);
+  request.count = htobe32 (512);
+  request.flags = htobe16 (0);
+  if (send (sock, &request, sizeof request, 0) != sizeof request) {
+    perror ("send: NBD_CMD_CACHE");
+    exit (EXIT_FAILURE);
+  }
+  if (recv (sock, &reply, sizeof reply, MSG_WAITALL) != sizeof reply) {
+    perror ("recv: NBD_CMD_CACHE");
+    exit (EXIT_FAILURE);
+  }
+  if (reply.error != NBD_SUCCESS) {
+    fprintf (stderr, "%s: NBD_CMD_CACHE failed with %d\n",
+             program_name, reply.error);
+    exit (EXIT_FAILURE);
+  }
+
+  short_sleep ();
+  log_verify_seen_in_order
+    ("testlayersfilter3: cache count=512 offset=0",
+     "filter3: test_layers_filter_cache",
+     "testlayersfilter2: cache count=512 offset=0",
+     "filter2: test_layers_filter_cache",
+     "testlayersfilter1: cache count=512 offset=0",
+     "filter1: test_layers_filter_cache",
+     "testlayersplugin: cache count=512 offset=0",
+     "test_layers_plugin_cache",
+     NULL);
+
   /* XXX We should test NBD_CMD_BLOCK_STATUS here.  However it
    * requires that we negotiate structured replies and base:allocation
    * in the handshake, and the format of the reply is more complex
@@ -549,14 +595,15 @@ main (int argc, char *argv[])
     perror ("waitpid");
 
   /* finalize methods called in reverse order of prepare */
-  sleep (1);
+  short_sleep ();
   log_verify_seen_in_order
     ("filter3: test_layers_filter_finalize",
      "filter2: test_layers_filter_finalize",
      "filter1: test_layers_filter_finalize",
      NULL);
 
-  /* close methods called in order */
+  /* close methods called outer-to-inner, which is reverse of completion
+   * of open */
   log_verify_seen_in_order
     ("filter3: test_layers_filter_close",
      "filter2: test_layers_filter_close",
@@ -576,7 +623,7 @@ main (int argc, char *argv[])
 }
 
 /* The log from nbdkit is captured in a separate thread. */
-static char *log = NULL;
+static char *log_buf = NULL;
 static size_t log_len = 0;
 static pthread_mutex_t log_lock = PTHREAD_MUTEX_INITIALIZER;
 
@@ -588,18 +635,19 @@ start_log_capture (void *arg)
   ssize_t r;
 
   for (;;) {
-    pthread_mutex_lock (&log_lock);
-    if (allocated <= log_len) {
-      allocated += 4096;
-      log = realloc (log, allocated);
-      if (log == NULL) {
-        perror ("log: realloc");
-        exit (EXIT_FAILURE);
+    {
+      ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&log_lock);
+      if (allocated <= log_len) {
+        allocated += 4096;
+        log_buf = realloc (log_buf, allocated);
+        if (log_buf == NULL) {
+          perror ("log: realloc");
+          exit (EXIT_FAILURE);
+        }
       }
     }
-    pthread_mutex_unlock (&log_lock);
 
-    r = read (fd, &log[log_len], allocated-log_len);
+    r = read (fd, &log_buf[log_len], allocated-log_len);
     if (r == -1) {
       perror ("log: read");
       exit (EXIT_FAILURE);
@@ -608,12 +656,11 @@ start_log_capture (void *arg)
       break;
 
     /* Dump the log as we receive it to stderr, for debugging. */
-    if (write (2, &log[log_len], r) == -1)
+    if (write (2, &log_buf[log_len], r) == -1)
       perror ("log: write");
 
-    pthread_mutex_lock (&log_lock);
+    ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&log_lock);
     log_len += r;
-    pthread_mutex_unlock (&log_lock);
   }
 
   /* nbdkit closed the connection. */
@@ -639,10 +686,9 @@ no_message_error (const char *msg)
 static void
 log_verify_seen (const char *msg)
 {
-  pthread_mutex_lock (&log_lock);
-  if (memmem (log, log_len, msg, strlen (msg)) == NULL)
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&log_lock);
+  if (memmem (log_buf, log_len, msg, strlen (msg)) == NULL)
     no_message_error (msg);
-  pthread_mutex_unlock (&log_lock);
 }
 
 static void messages_out_of_order (const char *msg1, const char *msg2)
@@ -663,31 +709,28 @@ log_verify_seen_in_order (const char *msg, ...)
   void *prev, *curr;
   const char *prev_msg, *curr_msg;
 
-  pthread_mutex_lock (&log_lock);
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&log_lock);
 
-  prev = memmem (log, log_len, msg, strlen (msg));
+  prev = memmem (log_buf, log_len, msg, strlen (msg));
   if (prev == NULL) no_message_error (msg);
   prev_msg = msg;
 
   va_start (args, msg);
   while ((curr_msg = va_arg (args, char *)) != NULL) {
-    curr = memmem (log, log_len, curr_msg, strlen (curr_msg));
+    curr = memmem (log_buf, log_len, curr_msg, strlen (curr_msg));
     if (curr == NULL) no_message_error (curr_msg);
     if (prev > curr) messages_out_of_order (prev_msg, curr_msg);
     prev_msg = curr_msg;
     prev = curr;
   }
   va_end (args);
-
-  pthread_mutex_unlock (&log_lock);
 }
 
 static void
 log_free (void)
 {
-  pthread_mutex_lock (&log_lock);
-  free (log);
-  log = NULL;
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&log_lock);
+  free (log_buf);
+  log_buf = NULL;
   log_len = 0;
-  pthread_mutex_unlock (&log_lock);
 }
