@@ -30,13 +30,14 @@
 # OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 # SUCH DAMAGE.
 
-# Test the memory plugin with the largest possible size supported
-# by nbdkit.
+# Test the memory plugin with the largest possible size supported by
+# nbdkit.  By using nbdsh (libnbd) as the client we are able to read
+# the final 511 byte "sector" which eludes qemu.
 
 source ./functions.sh
 set -e
 
-requires qemu-io --version
+requires nbdsh --version
 
 sock=`mktemp -u`
 files="memory-largest.out memory-largest.pid $sock"
@@ -45,13 +46,23 @@ cleanup_fn rm -f $files
 
 # Run nbdkit with memory plugin.
 # size = 2^63-1
-start_nbdkit -P memory-largest.pid -U $sock \
-       memory size=9223372036854775807
+start_nbdkit -P memory-largest.pid -U $sock memory 9223372036854775807
 
-# qemu cannot open this image!
-#
-#   can't open device nbd+unix://?socket=$sock: Could not get image size: File too large
-#
-# Therefore we skip the remainder of this test (in effect, testing
-# only that nbdkit can create the file).
-exit 77
+nbdsh --connect "nbd+unix://?socket=$sock" \
+      -c '
+# Write some stuff to the beginning, middle and end.
+buf1 = b"1" * 512
+h.pwrite (buf1, 0)
+buf2 = b"2" * 65536
+h.pwrite (buf2, 1000000001)
+buf3 = b"3" * 511
+h.pwrite (buf3, 9223372036854775296)
+
+# Read it back.
+buf11 = h.pread (len(buf1), 0)
+assert buf1 == buf11
+buf22 = h.pread (len(buf2), 1000000001)
+assert buf2 == buf22
+buf33 = h.pread (len(buf3), 9223372036854775296)
+assert buf3 == buf33
+'
