@@ -1,7 +1,6 @@
 #!/usr/bin/env bash
 # nbdkit
 # Copyright (C) 2017 Red Hat Inc.
-# All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions are
@@ -31,8 +30,10 @@
 # OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 # SUCH DAMAGE.
 
+source ./functions.sh
+
 pidfile=shebang-python.pid
-sockfile=shebang-python.sock
+sock=`mktemp -u`
 script=$SRCDIR/shebang.py
 
 if test ! -d "$SRCDIR"; then
@@ -40,9 +41,10 @@ if test ! -d "$SRCDIR"; then
     exit 1
 fi
 
-rm -f $pidfile $sockfile
+rm -f $pidfile $sock
+cleanup_fn rm -f $pidfile $sock
 
-$script -P $pidfile -U $sockfile -f -v &
+$script -P $pidfile -U $sock -f -v &
 
 # We may have to wait a short time for the pid file to appear.
 for i in {1..60}; do
@@ -57,12 +59,13 @@ if ! test -f $pidfile; then
 fi
 
 pid="$(cat $pidfile)"
+cleanup_fn kill $pid
 
 # Check the process exists.
 kill -s 0 $pid
 
 # Check the socket was created (and is a socket).
-test -S $sockfile
+test -S $sock
 
 # Kill the process.
 kill $pid
@@ -78,5 +81,3 @@ if kill -s 0 $pid; then
     echo "$0: process did not exit after sending a signal"
     exit 1
 fi
-
-rm $pidfile $sockfile
