@@ -1,7 +1,6 @@
 #!/usr/bin/env bash
 # nbdkit
 # Copyright (C) 2014 Red Hat Inc.
-# All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions are
@@ -35,22 +34,16 @@ source ./functions.sh
 set -e
 set -x
 
-if ! socat -h; then
-    echo "$0: 'socat' command not available"
-    exit 77
-fi
+requires socat -h
+requires qemu-img --version
 
-if ! qemu-img --help >/dev/null; then
-    echo "$0: 'qemu-img' command not available"
-    exit 77
-fi
-
-files="single.sock single.disk"
+sock=`mktemp -u`
+files="$sock single.disk"
 rm -f $files
 
 truncate -s 1G single.disk
 
-socat unix-listen:single.sock,reuseaddr,fork \
+socat unix-listen:$sock,reuseaddr,fork \
     exec:'nbdkit -r -s file single.disk' &
 pid=$!
 
@@ -63,14 +56,14 @@ cleanup_fn cleanup
 
 # Wait for socat to start up and create the socket.
 for i in {1..10}; do
-    if test -S single.sock; then
+    if test -S $sock; then
         break
     fi
     sleep 1
 done
-if ! test -S single.sock; then
+if ! test -S $sock; then
     echo "$0: socket was not created"
     exit 1
 fi
 
-qemu-img info nbd:unix:single.sock
+qemu-img info nbd:unix:$sock
