@@ -1,7 +1,6 @@
 #!/usr/bin/env bash
 # nbdkit
 # Copyright (C) 2018 Red Hat Inc.
-# All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions are
@@ -31,10 +30,12 @@
 # OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 # SUCH DAMAGE.
 
-set -x
+source ./functions.sh
 set -e
+set -x
 
 rm -f test-vddk.out
+cleanup_fn rm -f test-vddk.out
 
 LD_LIBRARY_PATH=.libs:$LD_LIBRARY_PATH \
 LIBRARY_PATH=.libs:$LIBRARY_PATH \
@@ -42,5 +43,3 @@ nbdkit vddk --dump-plugin > test-vddk.out
 cat test-vddk.out
 
 grep ^vddk_default_libdir= test-vddk.out
-
-rm test-vddk.out
