@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2018 Red Hat Inc.
+ * Copyright (C) 2018-2019 Red Hat Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
@@ -50,6 +50,11 @@
 
 #include <pthread.h>
 
+#ifndef SOCK_CLOEXEC
+/* For this file, we don't care if fds are marked cloexec; leaking is okay.  */
+#define SOCK_CLOEXEC 0
+#endif
+
 static char tmpdir[]   = "/tmp/wsXXXXXX";
 static char sockpath[] = "............./sock";
 static int listen_sock = -1;
@@ -61,7 +66,7 @@ static void *start_web_server (void *arg);
 static void handle_requests (int s);
 static void handle_request (int s, bool headers_only);
 static void xwrite (int s, const char *buf, size_t len);
-static void xpread (int fd, char *buf, size_t count, off_t offset);
+static void xpread (char *buf, size_t count, off_t offset);
 
 static void
 cleanup (void)
@@ -276,7 +281,7 @@ handle_request (int s, bool headers_only)
     exit (EXIT_FAILURE);
   }
 
-  xpread (fd, data, length, offset);
+  xpread (data, length, offset);
   xwrite (s, data, length);
 
   free (data);
@@ -299,7 +304,7 @@ xwrite (int s, const char *buf, size_t len)
 }
 
 static void
-xpread (int fd, char *buf, size_t count, off_t offset)
+xpread (char *buf, size_t count, off_t offset)
 {
   ssize_t r;
 
